from functools import cached_property
import itertools
from io import BufferedReader
import logging
import os
import re

from dreem.util.cmd import SAMTOOLS_CMD, run_cmd
from dreem.util.dflt import BUFFER_LENGTH
from dreem.util.seq import FastaParser
from dreem.util.path import BasePath, FastaInPath, XamTempPath

from functools import cached_property
import itertools
import logging
import os

from typing import Tuple, Dict

from dreem.util.cmd import FASTQC_CMD, CUTADAPT_CMD, BOWTIE2_CMD, \
    BOWTIE2_BUILD_CMD, run_cmd
from dreem.util.dflt import NUM_PROCESSES, PHRED_ENCODING
from dreem.util.cli import DEFAULT_MIN_BASE_QUALITY, DEFAULT_ILLUMINA_ADAPTER, DEFAULT_MIN_OVERLAP, DEFAULT_MAX_ERROR, DEFAULT_INDELS, DEFAULT_NEXTSEQ_TRIM, DEFAULT_DISCARD_TRIMMED, DEFAULT_DISCARD_UNTRIMMED, DEFAULT_MIN_LENGTH, DEFAULT_SCORE_MIN
from dreem.util.cli import DEFAULT_LOCAL, DEFAULT_UNALIGNED, DEFAULT_DISCORDANT, DEFAULT_MIXED, DEFAULT_DOVETAIL, DEFAULT_CONTAIN, DEFAULT_FRAG_LEN_MIN, DEFAULT_FRAG_LEN_MAX, DEFAULT_N_CEILING, DEFAULT_SEED_INTERVAL, DEFAULT_GAP_BAR, DEFAULT_SEED_SIZE, DEFAULT_EXTENSIONS, DEFAULT_RESEED, DEFAULT_PADDING, DEFAULT_ALIGN_THREADS, MATCH_BONUS, MISMATCH_PENALTY, N_PENALTY, REF_GAP_PENALTY, READ_GAP_PENALTY, IGNORE_QUALS
from dreem.util.path import SampleTempPath, FastaInPath, XamTempPath, TEMP_DIR, MOD_ALN, MOD_VEC, ALN_TRIM, ALN_ALIGN, ALN_REM, ALN_SORT, ALN_SPLIT, VEC_SELECT, VEC_SORT
from dreem.util.path import FastqInPath, Fastq1InPath, Fastq2InPath, FastqDemultiPath, Fastq1DemultiPath, Fastq2DemultiPath, SampleOutPath, XamTempPath, XamInPath, XamOutPath, XamIndexInPath, XamIndexTempPath, XamIndexSegment
from dreem.util.path import FastqOutPath, Fastq1OutPath, Fastq2OutPath, SAM_EXT, BAM_EXT, OUTPUT_DIR, XamSegment, PathError


# General parameters
DEFAULT_INTERLEAVED = False
SAM_HEADER = b"@"
SAM_ALIGN_SCORE = b"AS:i:"
SAM_EXTRA_SCORE = b"XS:i:"
FASTQ_REC_LENGTH = 4
DEFAULT_MIN_MAPQ = 30

# FastQC parameters
DEFAULT_EXTRACT = False


def _get_demultiplexed_mates(sample_path: SampleOutPath):
    mates1 = dict()
    mates2 = dict()
    for fq_file in os.listdir(sample_path.path):
        fq_path = str(sample_path.path.joinpath(fq_file))
        try:
            fq1 = Fastq1DemultiPath.parse(fq_path)
        except PathError:
            fq1 = None
        try:
            fq2 = Fastq2DemultiPath.parse(fq_path)
        except PathError:
            fq2 = None
        if fq1 and fq2:
            raise ValueError("Failed to determine whether FASTQ is mate 1 or 2:"
                             f" '{fq_path}'")
        if fq1:
            ref = fq1.fastq1.name
            if ref in mates1:
                raise ValueError(f"More than one FASTQ file for ref '{ref}'")
            mates1[ref] = fq1
        elif fq2:
            ref = fq2.fastq2.name
            if ref in mates2:
                raise ValueError(f"More than one FASTQ file for ref '{ref}'")
            mates2[ref] = fq2
        else:
            raise ValueError("FASTQ file name did not match any valid formats:"
                             f" '{fq_path}'")
    return mates1, mates2


def _get_mate_pairs_by_ref(mates1: Dict[str, Fastq1DemultiPath],
                           mates2: Dict[str, Fastq2DemultiPath]):
    refs1 = set(mates1)
    refs2 = set(mates2)
    refs = refs1 | refs2
    if (missing := refs - (refs1 & refs2)):
        raise ValueError("Missing mate 1 or 2 for refs "
                         f"{', '.join(missing)}")
    pairs: Dict[bytes, FastqUnit] = dict()
    for ref in refs:
        mate1 = str(mates1[ref].path)
        mate2 = str(mates2[ref].path)
        pairs[ref] = FastqUnit(fastq1=mate1, fastq2=mate2,
                               output=True, demultiplexed=True)
    return pairs


def get_demultiplexed_fastq_pairs(fq_dir: str):
    sample_path = SampleOutPath.parse(fq_dir)
    mates1, mates2 = _get_demultiplexed_mates(sample_path)
    pairs = _get_mate_pairs_by_ref(mates1, mates2)
    return pairs


class FastqUnit(object):
    _dm_classes = (FastqDemultiPath, FastqDemultiPath,
                   Fastq1DemultiPath, Fastq2DemultiPath)
    _out_classes = (FastqOutPath, FastqOutPath, Fastq1OutPath, Fastq2OutPath)
    _in_classes = (FastqInPath, FastqInPath, Fastq1InPath, Fastq2InPath)

    def __init__(self,
                 fastqu: str = "", fastqi: str = "",
                 fastq1: str = "", fastq2: str = "",
                 output: bool = False, demultiplexed: bool = False,
                 phred_encoding: int = PHRED_ENCODING):
        self._fqu = fastqu
        self._fqi = fastqi
        self._fq1 = fastq1
        self._fq2 = fastq2
        self._out = output
        self._demulti = demultiplexed
        self._phred = phred_encoding
    
    @property
    def _fq_paths(self):
        return {
            "fastqu": self._fqu,
            "fastqi": self._fqi,
            "fastq1": self._fq1,
            "fastq2": self._fq2,
        }

    @property
    def kwargs(self):
        return {
            **self._fq_paths,
            "output": self._out,
            "demultiplexed": self._demulti,
            "phred_encoding": self._phred,
        }

    @property
    def phred(self):
        return f"--phred{self._phred}"
    
    def _get_in_classes(self):
        if self._demulti:
            return self._dm_classes
        if self._out:
            return self._out_classes
        return self._in_classes
    
    def _get_out_classes(self):
        return self._dm_classes if self._demulti else self._out_classes
    
    @cached_property
    def _flags(self):
        is_unpaired = bool(self._fqu)
        is_interleaved = bool(self._fqi)
        has_paired1 = bool(self._fq1)
        has_paired2 = bool(self._fq2)
        if has_paired1 != has_paired2:
            raise ValueError("Cannot give fastq1 or fastq2 without the other")
        if (count := is_unpaired + is_interleaved + has_paired1) != 1:
            if count:
                raise ValueError("Too many FASTQ files given")
            raise ValueError("No FASTQ files given")
        return is_unpaired, is_interleaved, has_paired1, has_paired2
        
    @property
    def paired(self):
        return not self._flags[0]
    
    @property
    def interleaved(self):
        return self._flags[1]
    
    @property
    def two_files(self):
        return self._flags[2]
    
    @property
    def _in_classes_filtered(self):
        return tuple(cls for cls, flag in zip(self._get_in_classes(),
                                              self._flags, strict=True)
                     if flag)

    @property
    def _out_classes_filtered(self):
        return tuple(cls for cls, flag in zip(self._get_out_classes(),
                                              self._flags, strict=True)
                     if flag)
    
    @property
    def _fq_items_filtered(self):
        return tuple(item for item, flag in zip(self._fq_paths.items(),
                                                self._flags, strict=True)
                     if flag)
    
    @property
    def _fq_args_filtered(self):
        return tuple(arg for arg, _ in self._fq_items_filtered)

    @property
    def _fq_paths_filtered(self):
        return tuple(path for _, path in self._fq_items_filtered)
    
    @property
    def fqs(self):
        return [cls.parse(path) for cls, path in
                zip(self._in_classes_filtered, self._fq_paths_filtered,
                    strict=True)]
    
    @property
    def paths(self):
        return [fq.path for fq in self.fqs]
    
    def _get_single_value(self, values: list, name: str):
        if not values:
            raise ValueError(f"No {name} for {self}")
        if len(values) > 2:
            raise ValueError(f"Too many {name}s for {self}")
        if len(values) == 2 and values[0] != values[1]:
            raise ValueError(f"Inconsistent {name}s for {self}")
        return values[0]

    @property
    def sample(self) -> str:
        return self._get_single_value(
            [fq.last.name for fq in self.fqs], "sample")
    
    @property
    def ref(self) -> bytes:
        if self._demulti:
            return self._get_single_value(
                [fq.last.name for fq in self.fqs], "ref")
        raise ValueError("Cannot get ref from non-demultiplexed FASTQs")
    
    @property
    def cutadapt_input_args(self):
        return self.paths
    
    @property
    def cutadapt_output_args(self):
        args = list()
        for flag, output in zip(("-o", "-p"), self.paths, strict=False):
            args.extend([flag, output])
        return args

    @property
    def bowtie2_input_args(self):
        unpaired, interleaved, mate1, mate2 = self._flags
        if unpaired:
            flags = ("-U",)
        elif interleaved:
            flags = ("--interleaved",)
        else:
            assert mate1 and mate2
            flags = ("-1", "-2")
        return list(itertools.chain(*zip(flags, self.paths, strict=True)))
    
    def rename(self, **kwargs: str):
        renamed = dict()
        for fq, fq_arg, cls in zip(self.fqs, self._fq_args_filtered,
                                   self._out_classes_filtered, strict=True):
            fq_kwargs = {**fq.kwargs, **kwargs}
            if self._demulti:
                fq_kwargs["sample"] = self.sample
            renamed[fq_arg] = cls(**fq_kwargs).path
        return self.__class__(**renamed)


class ReadsFileBase(object):
    partition = TEMP_DIR
    module = MOD_ALN
    _step = ""

    def __init__(self, base_path: BasePath) -> None:
        self.base_path = base_path
    
    @property
    def step(self):
        if self._step:
            return self._step
        raise NotImplementedError
    
    def setup(self):
        raise NotImplementedError

    def run(self):
        raise NotImplementedError
    
    def clean(self):
        raise NotImplementedError


class FastqBase(ReadsFileBase):
    def __init__(self, base_path: BasePath, fq_unit: FastqUnit) -> None:
        super().__init__(base_path)
        self.fq_unit = fq_unit
    
    @property
    def paired(self):
        return self.fq_unit.paired

    @property
    def _path_args(self):
        return {
            "base": self.base_path,
            "partition": self.partition,
            "module": self.module,
            "temp_step": self.step,
            "sample": self.fq_unit.sample
        }
    
    @property
    def output_dir(self):
        return SampleTempPath(**self._path_args)
    
    def qc(self, extract: bool = DEFAULT_EXTRACT):
        cmd = [FASTQC_CMD]
        if extract:
            cmd.append("--extract")
        cmd.extend(self.fq_unit.paths)
        run_cmd(cmd)

    def setup(self):
        self.output_dir.path.mkdir(parents=True, exist_ok=True)


class FastqTrimmer(FastqBase):
    _step = ALN_TRIM

    @property
    def output(self):
        return self.fq_unit.rename(**self._path_args)

    def _cutadapt(self,
                  qual1: int=DEFAULT_MIN_BASE_QUALITY,
                  qual2: int=0,
                  adapters15: Tuple[str]=(),
                  adapters13: Tuple[str]=(DEFAULT_ILLUMINA_ADAPTER,),
                  adapters25: Tuple[str]=(),
                  adapters23: Tuple[str]=(DEFAULT_ILLUMINA_ADAPTER,),
                  min_overlap: int=DEFAULT_MIN_OVERLAP,
                  max_error: float=DEFAULT_MAX_ERROR,
                  indels: bool=DEFAULT_INDELS,
                  nextseq_trim: bool=DEFAULT_NEXTSEQ_TRIM,
                  discard_trimmed: bool=DEFAULT_DISCARD_TRIMMED,
                  discard_untrimmed: bool=DEFAULT_DISCARD_UNTRIMMED,
                  min_length: bool=DEFAULT_MIN_LENGTH,
                  cores: int=NUM_PROCESSES):
        cmd = [CUTADAPT_CMD]
        if cores >= 0:
            cmd.extend(["--cores", cores])
        if nextseq_trim:
            if qual1 > 0:
                cmd.extend(["--nextseq-trim", qual1])
        else:
            if qual1 > 0:
                cmd.extend(["-q", qual1])
            if qual2 > 0:
                cmd.extend(["-Q", qual2])
        adapters = {"g": adapters15, "a": adapters13,
                    "G": adapters25, "A": adapters23}
        for arg, adapter in adapters.items():
            if adapter and (self.paired or arg.islower()):
                for adapt in adapter:
                    cmd.extend([f"-{arg}", adapt])
        if min_overlap >= 0:
            cmd.extend(["-O", min_overlap])
        if max_error >= 0:
            cmd.extend(["-e", max_error])
        if not indels:
            cmd.append("--no-indels")
        if discard_trimmed:
            cmd.append("--discard-trimmed")
        if discard_untrimmed:
            cmd.append("--discard-untrimmed")
        if min_length:
            cmd.extend(["-m", min_length])
        cmd.extend(["--report", "minimal"])
        if self.fq_unit.interleaved:
            cmd.append("--interleaved")
        cmd.extend(self.output.cutadapt_output_args)
        cmd.extend(self.fq_unit.cutadapt_input_args)
        run_cmd(cmd)
        return self.output
    
    def run(self, **kwargs):
        self.setup()
        return self._cutadapt(**kwargs)
    
    def clean(self):
        for path in self.output.paths:
            path.unlink()


class FastqAligner(FastqBase):
    _step = ALN_ALIGN

    def __init__(self, base_dir: BasePath,
                 fastq: BasePath, refs_file: FastaInPath) -> None:
        super().__init__(base_dir, fastq)
        self.refs_file = refs_file
    
    @property
    def refs_name(self):
        return self.refs_file.path.stem
    
    @property
    def refs_prefix(self):
        return self.refs_file.path.with_suffix("")

    @property
    def output(self):
        kwargs = {**self._path_args,
                  "xam": XamSegment.format(self.refs_name)}
        return XamTempPath(**kwargs)

    def _bowtie2_build(self):
        """
        Build an index of a reference genome using Bowtie 2.
        :param ref: (str) path to the reference genome FASTA file
        :return: None
        """
        cmd = [BOWTIE2_BUILD_CMD, "-q", self.refs_file, self.refs_prefix]
        run_cmd(cmd)
    
    def _bowtie2(self,
                 local=DEFAULT_LOCAL,
                 unaligned=DEFAULT_UNALIGNED,
                 discordant=DEFAULT_DISCORDANT,
                 mixed=DEFAULT_MIXED,
                 dovetail=DEFAULT_DOVETAIL,
                 contain=DEFAULT_CONTAIN,
                 score_min=DEFAULT_SCORE_MIN,
                 frag_len_min=DEFAULT_FRAG_LEN_MIN,
                 frag_len_max=DEFAULT_FRAG_LEN_MAX,
                 n_ceil=DEFAULT_N_CEILING,
                 gap_bar=DEFAULT_GAP_BAR,
                 seed_size=DEFAULT_SEED_SIZE,
                 seed_interval=DEFAULT_SEED_INTERVAL,
                 extensions=DEFAULT_EXTENSIONS,
                 reseed=DEFAULT_RESEED,
                 padding=DEFAULT_PADDING,
                 threads=DEFAULT_ALIGN_THREADS):
        cmd = [BOWTIE2_CMD]
        cmd.extend(self.fq_unit.bowtie2_input_args)
        cmd.extend(["-x", self.refs_prefix])
        cmd.extend(["-S", self.output])
        cmd.append(self.fq_unit.phred)
        cmd.append("--xeq")
        cmd.extend(["--ma", MATCH_BONUS])
        cmd.extend(["--mp", MISMATCH_PENALTY])
        cmd.extend(["--np", N_PENALTY])
        cmd.extend(["--rfg", REF_GAP_PENALTY])
        cmd.extend(["--rdg", READ_GAP_PENALTY])
        if score_min:
            cmd.extend(["--score-min", score_min])
        if local:
            cmd.append("--local")
        if not unaligned:
            cmd.append("--no-unal")
        if not discordant:
            cmd.append("--no-discordant")
        if not mixed:
            cmd.append("--no-mixed")
        if dovetail:
            cmd.append("--dovetail")
        if not contain:
            cmd.append("--no-contain")
        if frag_len_min:
            cmd.extend(["-I", frag_len_min])
        if frag_len_max:
            cmd.extend(["-X", frag_len_max])
        if n_ceil:
            cmd.extend(["--n-ceil", n_ceil])
        if gap_bar:
            cmd.extend(["--gbar", gap_bar])
        if seed_size:
            cmd.extend(["-L", seed_size])
        if seed_interval:
            cmd.extend(["-i", seed_interval])
        if extensions:
            cmd.extend(["-D", extensions])
        if reseed:
            cmd.extend(["-R", reseed])
        if padding:
            cmd.extend(["--dpad", padding])
        if threads:
            cmd.extend(["-p", threads])
        if IGNORE_QUALS:
            cmd.append("--ignore-quals")
        run_cmd(cmd)
        return self.output
    
    def run(self, **kwargs):
        self.setup()
        self._bowtie2_build()
        return self._bowtie2(**kwargs)
    
    def clean(self):
        self.output.path.unlink()


class XamBase(ReadsFileBase):
    ext = SAM_EXT

    def __init__(self, base_path: BasePath, xam_in: XamInPath | XamTempPath) -> None:
        super().__init__(base_path)
        self.xam_in = xam_in
    
    @property
    def xai_class(self):
        if isinstance(self.xam_in, XamInPath):
            return XamIndexInPath
        if isinstance(self.xam_in, XamTempPath):
            return XamIndexTempPath
        raise TypeError(self.xam_in)
    
    @classmethod
    def out_class(cls):
        if cls.partition == TEMP_DIR:
            return XamTempPath
        if cls.partition == OUTPUT_DIR:
            return XamOutPath
        raise NotImplementedError
    
    @property
    def sample(self):
        return self.xam_in.sample

    @property
    def name_in(self):
        return self.xam_in.xam.name
    
    @property
    def name_out(self):
        return self.xam_out.xam.name
    
    def _get_xam_out(self, name: str):
        kwargs = {
            "base": self.base_path,
            "partition": self.partition,
            "module": self.module,
            "sample": self.sample
        }
        if self.partition == TEMP_DIR:
            kwargs["temp_step"] = self.step
        kwargs["xam"] = XamSegment.format(name, ext=self.ext)
        return self.out_class()(**kwargs)
    
    @cached_property
    def xam_out(self):
        return self._get_xam_out(self.name_in)
    
    @property
    def output_dir(self):
        return self.xam_out.path.parent
    
    def _make_output_dir(self):
        self.output_dir.mkdir(parents=True, exist_ok=True)
    
    @property
    def xam_index(self):
        xai_seg = XamIndexSegment.format(self.xam_in.last.name)
        args = self.xam_in.args[:-1] + (xai_seg,)
        return self.xai_class(*args)

    def create_index(self):
        cmd = [SAMTOOLS_CMD, "index", self.xam_in]
        run_cmd(cmd)
        xam_index = self.xam_index
        assert xam_index.path.is_file()
        return xam_index
    
    def clean(self):
        self.xam_out.path.unlink()


class SamRemoveEqualMappers(XamBase):
    step = ALN_REM

    pattern_a = re.compile(SAM_ALIGN_SCORE + rb"(\d+)")
    pattern_x = re.compile(SAM_EXTRA_SCORE + rb"(\d+)")

    @staticmethod
    def _get_score(line: bytes, ptn: re.Pattern[bytes]):
        return (float(match.groups()[0])
                if (match := ptn.search(line)) else None)
    
    @classmethod
    def _is_best_alignment(cls, line: bytes):
        return ((score_x := cls._get_score(line, cls.pattern_x)) is None
                or score_x < cls._get_score(line, cls.pattern_a))
    
    @staticmethod
    def _read_is_paired(line: bytes):
        info = line.split()
        if len(info) < 11:  # FIXME: define magic number
            raise ValueError(f"Invalid SAM line:\n{line.decode()}")
        flag = int(info[1])
        if 0 <= flag < 2**12:  # FIXME: define magic number 
            return bool(flag % 2)
        raise ValueError(f"Invalid SAM flag: {flag}")
    
    def _iter_paired(self, sam: BufferedReader, line: bytes):
        for line2 in sam:
            if self._is_best_alignment(line) or self._is_best_alignment(line2):
                yield b"".join((line, line2))
            line = sam.readline()
    
    def _iter_single(self, sam: BufferedReader, line: bytes):
        while line:
            if self._is_best_alignment(line):
                yield(line)
            line = sam.readline()

    def _remove_equal_mappers(self, buffer_length=BUFFER_LENGTH):
        with (open(self.xam_in.path, "rb") as sami,
              open(self.xam_out.path, "wb") as samo):
            # Copy the header from the input to the output SAM file.
            while (line := sami.readline()).startswith(SAM_HEADER):
                samo.write(line)
            if line:
                paired = self._read_is_paired(line)
                iter_sam = self._iter_paired if paired else self._iter_single
                lines = iter_sam(sami, line)
                while text := b"".join(itertools.islice(lines, buffer_length)):
                    samo.write(text)
        return self.xam_out
    
    def run(self):
        logging.info("\nRemoving Reads Mapping Equally to Multiple Locations"
                     f" in {self.xam_in}\n")
        self._make_output_dir()
        return self._remove_equal_mappers()


class XamSorter(XamBase):
    def _sort(self, name: bool):
        cmd = [SAMTOOLS_CMD, "sort"]
        if name:
            cmd.append("-n")
        cmd.extend(["-o", self.xam_out, self.xam_in])
        run_cmd(cmd)
        return self.xam_out
    
    def run(self, name: bool = False):
        logging.info(f"\nSorting {self.xam_in} by Reference and Coordinate\n")
        self._make_output_dir()
        return self._sort(name)


class BamAlignSorter(XamSorter):
    step = ALN_SORT
    ext = BAM_EXT


class SamVectorSorter(XamSorter):
    module = MOD_VEC
    step = VEC_SORT
    ext = SAM_EXT


class BamSplitter(XamBase):
    partition = OUTPUT_DIR
    step = ALN_SPLIT
    ext = BAM_EXT

    def __init__(self, base_path: BasePath,
                 xam_in: XamInPath | XamTempPath,
                 refs_file: FastaInPath) -> None:
        super().__init__(base_path, xam_in)
        self._refs_file = refs_file
    
    @cached_property
    def refs(self):
        return [ref for ref, _ in FastaParser(self._refs_file.path).parse()]
    
    def _extract_one_ref(self, ref: bytes):
        xam_out = self._get_xam_out(ref)
        cmd = [SAMTOOLS_CMD, "view", "-b", "-o", xam_out, self.xam_in, ref]
        run_cmd(cmd)
        return xam_out
    
    def _split_bam(self):
        self.create_index()
        return list(map(self._extract_one_ref, self.refs)) 
    
    def run(self):
        logging.info(f"\nSplitting {self.xam_in} into Individual References\n")
        self._make_output_dir()
        return self._split_bam()


class BamVectorSelector(XamBase):
    module = MOD_VEC
    step = VEC_SELECT
    ext = BAM_EXT
    
    @staticmethod
    def ref_coords(ref: str, first: int, last: int):
        return f"{ref}:{first}-{last}"

    def _select(self, ref: str, first: int, last: int):
        cmd = [SAMTOOLS_CMD, "view", "-h", "-o", self.xam_out, self.xam_in,
               self.ref_coords(ref, first, last)]
        run_cmd(cmd)
        return self.xam_out
    
    def run(self, ref: str, first: int, last: int):
        logging.info(f"\nSelecting {self.xam_in} ref {ref}: {first}-{last}\n")
        self._make_output_dir()
        return self._select(ref, first, last)



'''
primer1 = "CAGCACTCAGAGCTAATACGACTCACTATA"
primer1rc = "TATAGTGAGTCGTATTAGCTCTGAGTGCTG"
primer2 = "TGAAGAGCTGGAACGCTTCACTGA"
primer2rc = "TCAGTGAAGCGTTCCAGCTCTTCA"
adapters5 = (primer1, primer2rc)
adapters3 = (primer2, primer1rc)
'''



import numpy as np
import scipy.stats
from scipy.optimize import newton_krylov

from dreem.util.cli import *

## ------------- Utility functions ------------- ##

def calc_BIC(N, PARAMS_LEN, K, log_like):
    """
    """
    return np.log(N) * PARAMS_LEN * K - (2 * log_like)

def calc_denom(i, mu, denom_probs, s2_probs):
    """
    """
    mu = np.clip(mu, 0.0, 1.0)
    if i in denom_probs:  # Already encountered
        return (denom_probs[i], s2_probs)
    elif i >= len(mu):  # Base case
        return (1, s2_probs)
    else:  # Make the calc
        s1 = calc_denom(i + 1, mu, denom_probs, s2_probs)[0]
        s2 = (1.0 - mu[i + 1: i + 4]).prod() * \
            calc_denom(i + 4, mu, denom_probs, s2_probs)[0]
        denom_probs[i] = ((1 - mu[i]) * s1) + (mu[i] * s2)
        s2_probs[i] = s2
        return (denom_probs[i], s2_probs)

def mu_der(mu_k, x_bar_k):
    """
    """
    mu_k_rev = mu_k[::-1]
    denom_k = calc_denom(0, mu_k, {}, {})
    denom_k_rev = calc_denom(0, mu_k_rev, {}, {})
    upd_mu = [(mu_k[i] * denom_k[1][i] * denom_k_rev[1][len(mu_k) - i - 1] /
              denom_k[0]) - x_bar_k[i]
              for i in range(len(mu_k))]
    return np.array(upd_mu)

    

class EMclustering:
    """This class runs the EM clustering algorithm.
    
    Parameters from args:
    ---------------------
    
    bv: array (N x D)
        Bitvector of the construct. 
    
    K: int
        Number of clusters.
        
    read_hist: array (N)
        Number of reads per bitvector.
        
    min_iter: int
        Minimum number of iterations.
        
    convergence_eps: float
        Convergence threshold. When the difference between the log-likelihood of two consecutive iterations is lower than convergence_eps, the algorithm stops.
        
    Key methods:
    ------------
    
    run: Run the EM clustering algorithm. 
    maximize: Maximization step of the EM algorithm.
    expectation: Expectation step of the EM algorithm.
    
    """
    
    def __init__(self, bv, K, read_hist, bases_to_keep, sequence, n_cpus:int=N_CPUS, max_clusters:int=MAX_CLUSTERS, signal_thresh:float=SIGNAL_THRESH, info_thresh:float=INFO_THRESH, include_g_u:bool=INCLUDE_G_U, include_del:bool=INCLUDE_DEL, min_reads:int=MIN_READS, min_iter:int=MIN_ITER, convergence_cutoff:float=CONVERGENCE_CUTOFF, num_runs:int=NUM_RUNS, verbose:bool=VERBOSE):
        self.bv = bv
        self.sparse_mu = np.zeros((K, len(sequence)))
        self.bases_to_keep = bases_to_keep
        self.K = K
        self.read_hist = read_hist
        self.N = bv.shape[0]
        self.D = bv.shape[1]
        
        self.min_iter = min_iter
        self.cpus = n_cpus

        self.convergence_eps = convergence_cutoff
    
    def expectation(self, mu, pi):
        """
        Run the Expectation step of the EM algorithm - calc log like
        Args:
            X (EM_Class object): Contains the list of bit vectors
            K (int): Number of clusters
            mu (list): DMS reactivities in each cluster
            pi (list): Proportion of each cluster
            calc_inds (list): Indices to split the matrix at
            n_cpus (int): Number of cpus to use
            max_clusters (int): Maximum number of clusters
            signal_thresh (float): Threshold for signal
            info_thresh (float): Threshold for information
            include_g_u (bool): Include G and U in the calculation
            include_del (bool): Include deletions in the calculation
            min_reads (int): Minimum number of reads for a cluster
            min_iter (int): Minimum number of iterations
            convergence_cutoff (float): Convergence threshold
            num_runs (int): Number of runs
            verbose (bool): Print verbose output
            
        Returns:
            resps (list): Responsibity of each cluster
            log_like (float): Log likelihood of observing the data
            denom (float): Corrected denom of EM algorithm
        """
        N, D = self.bv.shape[0], self.bv.shape[1]
        log_pi = np.log(pi)
        log_pmf = np.zeros((N, D, self.K))
        for k in range(self.K):
            self.sparse_mu[k][self.bases_to_keep] = mu[k]
        denom = [calc_denom(0, self.sparse_mu[k], {}, {})[0] for k in range(self.K)]
        

        # Compute probability mass function for all reads, for each cluster
        for k in range(self.K):
            log_pmf[:, :, k] = scipy.stats.bernoulli.logpmf(self.bv, mu[k]) 

        log_pmf = np.sum(log_pmf, axis=1)  # Sum of log - like taking product

        # Substract log of denominator - like dividing by it
        for k in range(self.K):
            log_pmf[:, k] -= np.log(denom[k])

        log_resps_numer = log_pi + log_pmf
        log_resps_denom = scipy.special.logsumexp(log_resps_numer, axis=1)
        log_resps = (log_resps_numer.T - log_resps_denom).T
        resps = np.exp(log_resps)

        log_like = np.dot(log_resps_denom, self.read_hist)
        
        return (resps, log_like, denom)


    def maximisation(self, mu, resps, denom):
        """
        Run the Maximization step of the EM algorithm - update mu and pi
        Args:
            X (EM_Class object): Contains the list of bit vectors
            K (int): Number of clusters
            mu (list): DMS reactivities in each cluster
            resps (list): Responsibity of each cluster
            denom (float): Corrected denom of EM algorithm
        Returns:
            mu (list): DMS reactivities in each cluster
            obs_pi (list): Observed proportion of each cluster
            real_pi (list): Corrected proportion of each cluster
        """
        D = self.bv.shape[1]
        mu, obs_pi, real_pi = np.zeros((self.K, D)), np.zeros(self.K), np.zeros(self.K)
        for k in range(self.K):
            N_k = np.sum(resps[:, k] * self.read_hist)
            x_bar_k = np.sum((resps[:, k] * self.read_hist *
                            self.bv.T).T, axis=0) / N_k

            sparse_x_bar_k = np.zeros(self.sparse_mu.shape[1])
            sparse_x_bar_k[self.bases_to_keep] = x_bar_k
            self.sparse_mu[k][self.bases_to_keep] = mu[k]

            upd_mu = newton_krylov(lambda mu_k: mu_der(mu_k, sparse_x_bar_k), self.sparse_mu[k])
            upd_mu = upd_mu[self.bases_to_keep]

            # Check if mu is less than 0. This should not happen.
            zeros = np.where(upd_mu < 0.0)[0]
            if len(zeros) > 0:
                print('Mu is 0 or less:', upd_mu[upd_mu <= 0.0])
                upd_mu = np.maximum(upd_mu, np.zeros(len(upd_mu)))

            mu[k] = upd_mu  # Mu with denom correction
            # mu[k] = x_bar_k  # Mu without denom correction
            obs_pi[k] = N_k / np.sum(self.read_hist)
        real_pi = [obs_pi[k] / denom[k] for k in range(self.K)]
        real_pi = real_pi / np.sum(real_pi)
        
        return (mu, obs_pi, real_pi)


    def run(self):

        """Run the EM clustering algorithm.
        
        Output:
        -------
        A dictionary containing the following attributes:
            mu: array (K x D)
                Mean of each cluster.
            pi: array (K)
                Probability of each cluster.
            log_likelihood: float
                Log-likelihood of the model.
        
        """
        """
        Run the EM algorithm on DMSMaPseq data

        Returns:
            log_like_list (list): Log likelihoods
            final_mu (list): DMS reactivities in each cluster
            final_obs_pi (list): Obs proportion of each cluster
            final_real_pi (list): Corrected proportion of each cluster
            resps (list): Responsibity of each cluster
            BIC (float): BIC value for this K
        """

        N, D = self.bv.shape[0], self.bv.shape[1]

        # ---------------------- Iterations start ---------------------------- #

        # Initialize DMS modification rate for each base in each cluster
        # by sampling from a beta distribution
        BETA_A = 1.5  # Beta dist shape parameter
        BETA_B = 20  # Beta dist shape parameter
        mu = np.asarray([scipy.stats.beta.rvs(BETA_A, BETA_B, size=D)
                        for k in range(self.K)])
        # Initialize cluster probabilties with uniform distribution
        obs_pi = np.asarray([1.0 / self.K] * self.K)

        converged = False
        iter = 1
        log_like_list, mu_list, obs_pi_list, real_pi_list = [], [], [], []

        while not converged:  # Each iteration of the EM algorithm
            # print('Iteration:', iter, '|', np.min(mu), np.max(mu))

            # Expectation step
            (resps, log_like, denom) = self.expectation(mu, obs_pi)

            # Maximization step
            (mu, obs_pi, real_pi) = self.maximisation(mu, resps, denom)

            log_like_list.append(log_like)
            mu_list.append(mu)
            obs_pi_list.append(obs_pi)
            real_pi_list.append(real_pi)

            # Check if log like is decreasing - why does this happen?
            if iter > 1 and log_like < log_like_list[-2]:
                prev_loglike = log_like_list[-2]
                print('The log like decreased from {:.9f} to {:.9f}'.format(prev_loglike, log_like))
                if iter >= self.min_iter:
                    converged = True
            else:
                if iter >= self.min_iter:  # At least min iterations has run
                    prev_loglike = log_like_list[-2]
                    diff = log_like - prev_loglike
                    if diff <= self.convergence_eps:  # Converged
                        converged = True
                        print('Log like converged after {:d} iterations'.format(iter))
            iter += 1            

        final_mu, final_obs_pi, final_real_pi = mu_list[-1], obs_pi_list[-1], real_pi_list[-1]

        # ------------------------ Iterations end ---------------------------- #

        # BIC = calc_BIC(N, D, self.K, log_like_list[-1]) ## !! Technically it should be the number of non G/T bases, not D !!
        return {'mu': final_mu, 'pi': final_real_pi, 'log_likelihood': log_like_list[-1]}


## ----- Testing if the above code gives same results as original code ----- ##

if __name__ == '__main__' and False:

    exp_path = "/Users/Alberic/Desktop/Pro/RouskinLab/projects/DREEM/"
    bit_Vector_total = np.load(exp_path+"bit_vector.npy")
    read_hist_total =  np.load(exp_path+"read_hist.npy")

    # with open(exp_path+"data_EM_analysis.txt", 'a') as f:
    #     f.write("N_reads n_cpu dT mean_memory peak_memory \n")
    #     f.close()
    
    # for N_partial in np.geomspace(10000, 120000, 5).astype(np.int64):
    for N_partial in [10000]:

        N_partial = min(N_partial, bit_Vector_total.shape[0])

        # for n_cpu in range(1, 11):
        for n_cpu in [2]:

            print("Starting experiment with {} reads and {} cpus".format(N_partial, n_cpu))
            
            tracemalloc.start()

            # Get part of the total reads
            bit_Vector = copy.copy(bit_Vector_total[:N_partial])
            read_hist = copy.copy(read_hist_total[:N_partial])
            
            # Run EM and record stats
            EM = EMclustering(bit_Vector, K=2, read_hist=read_hist, min_iter=10, n_cpus=n_cpu,
                            max_clusters=3, signal_thresh=0.5, info_thresh=0.5, include_g_u=True, include_del=True,
                            min_reads=10,convergence_cutoff=0.5,num_runs=100, verbose=True )
            
            result = EM.run()

            # Log experiment stats
            # memory_stats = [ mem/1e6 for mem in tracemalloc.get_traced_memory() ]
            # log_data = np.array((N_partial,
            #                         n_cpu,
            #                         result["dT"],
            #                         memory_stats[0], 
            #                         memory_stats[1] ))
 
            # # with open(exp_path+"data_EM_analysis.txt", 'a') as f:
            #     np.savetxt(f, log_data, newline=" ")
            #     f.write("\n")
            #     f.close()

            # Stop memory tracing and clean memory
            tracemalloc.stop()
            bit_Vector = None
            read_hist = None
            # time.sleep(5)
            
            # Comparing output with previous software -> different test
            mu_reference = np.load("/Users/Alberic/Desktop/Pro/RouskinLab/projects/DREEM/result.npy")
            print("Reference matched:",np.allclose(mu_reference, result["mu"], atol=1e-9))

            for k in range(2):
                plt.subplot(1, 2, k+1)
                plt.plot(result["mu"][k], alpha=0.5)
                plt.plot(mu_reference[k], alpha=0.5)




from __future__ import annotations
import re
from typing import List, Optional

from dreem.util.util import BASES, SUB_A, SUB_C, SUB_G, SUB_T, SUB_N, MATCH, DELET, ANY_N, INS_3, INS_5, BLANK, DNA, DEFAULT_PHRED_ENCODING


CIG_ALN = b"M"  # alignment match
CIG_MAT = b"="  # sequence match
CIG_SUB = b"X"  # substitution
CIG_DEL = b"D"  # deletion
CIG_INS = b"I"  # insertion
CIG_SCL = b"S"  # soft clipping
CIG_PATTERN = re.compile(b"".join(
    [rb"(\d+)([", CIG_ALN, CIG_MAT, CIG_SUB, CIG_DEL, CIG_INS, CIG_SCL, b"])"]
))
SAM_HEADER = b"@"
NOLIM = -1
A_INT, C_INT, G_INT, T_INT = BASES
MATCH_INT = MATCH[0]
DELET_INT = DELET[0]
INS_5_INT = INS_5[0]
INS_3_INT = INS_3[0]
SUB_A_INT = SUB_A[0]
SUB_C_INT = SUB_C[0]
SUB_G_INT = SUB_G[0]
SUB_T_INT = SUB_T[0]
SUB_N_INT = SUB_N[0]
ANY_N_INT = ANY_N[0]
MIN_QUAL_PHRED = 20
MIN_QUAL_PCODE = MIN_QUAL_PHRED + DEFAULT_PHRED_ENCODING


def encode_base(base: int):
    if base == T_INT:
        return SUB_T_INT
    if base == G_INT:
        return SUB_G_INT
    if base == C_INT:
        return SUB_C_INT
    if base == A_INT:
        return SUB_A_INT
    raise ValueError(f"Invalid base: {base.to_bytes().decode()}")


def encode_compare(ref_base: int, read_base: int, read_qual: int):
    return ((MATCH_INT if ref_base == read_base else encode_base(read_base))
            if read_qual >= MIN_QUAL_PCODE
            else ANY_N_INT ^ encode_base(ref_base))


def encode_match(read_base: int, read_qual: int):
    # A more efficient version of encode_compare given the prior knowledge from
    # the CIGAR string that the read and reference match at this position.
    # NOTE: there is no analagous version when there is a known substitution
    # because substitutions are infrequent, so optimizing their processing
    # would speed the program insignificantly while making the source code
    # more complex and harder to maintain.
    return (MATCH_INT
            if read_qual >= MIN_QUAL_PCODE
            else ANY_N_INT ^ encode_base(read_base))


class Indel(object):
    """
    Base class for an Insertion or Deletion.

    It is used to find alternative positions for insertions and deletions,
    by keeping track of an indel's current coordinates (as they are moved)
    and determining whether a specific move is valid.

    Arguments
    rel_ins_idx (int):  The 0-indexed position of the indel with respect to
                        the sequence (ref or read) with a relative insertion
                        (that is, the read if the mutation is denoted an
                        insertion, and the ref if a deletion).
                        This index points to one specific base, either
                        (for insertions) a base inserted into the read,
                        in the coordinates of the read or (for deletions)
                        a base present in the ref and absent from the read,
                        in the coordinates of the ref.
    rel_del_idx (int):  The opposite of rel_ins_idx: the 0-indexed position
                        of the indel with respect to the sequence with a
                        relative deletion (that is, the read if the mutation
                        is denoted a deletion, and the ref if an insertion).
                        Because the deleted base does not actually exist
                        within the sequence whose coordinates it is based on,
                        rel_del_idx does not refer to a specific position in
                        the sequence, but instead to the two extant positions
                        in the sequence that flank the deleted position.
                        It is most convenient for the algorithm to have this
                        argument refer to the position 3' of the deleted base
                        and define the 5' position as a property.
    """

    __slots__ = ["_ins_idx", "_ins_init", "_del_idx", "_del_init", "_tunneled"]

    MIN_INDEL_DIST = 2

    def __init__(self, rel_ins_idx: int, rel_del_idx: int) -> None:
        self._ins_idx = rel_ins_idx
        self._ins_init = rel_ins_idx
        self._del_idx = rel_del_idx
        self._del_init = rel_del_idx
        self._tunneled = False

    @property
    def ins_idx(self):
        return self._ins_idx
    
    @property
    def del_idx5(self):
        return self._del_idx - 1

    @property
    def del_idx3(self):
        return self._del_idx
    
    @property
    def tunneled(self):
        return self._tunneled
    
    @property
    def rank(self) -> int:
        raise NotImplementedError

    def reset(self):
        self._ins_idx = self._ins_init
        self._del_idx = self._del_init
        self._tunneled = False
    
    @staticmethod
    def _get_indel_by_idx(indels: List[Indel], idx: int):
        for indel in indels:
            if indel.ins_idx == idx:
                return indel

    def _peek_out_of_indel(self, indels: List[Indel], from3to5: bool):
        inc = -1 if from3to5 else 1
        idx = self.ins_idx + inc
        tunneled_indels: List[Indel] = list()
        while indel := (self._get_indel_by_idx(indels, idx)):
            idx += inc
            tunneled_indels.append(indel)
        self._tunneled = bool(tunneled_indels)
        return idx, tunneled_indels
    
    def _collision(self, other: Indel, swap_idx: int):
        return self.MIN_INDEL_DIST > (min(abs(swap_idx - other.del_idx5),
                                          abs(swap_idx - other.del_idx3)))
    
    def _collisions(self, indels: List[Indel], swap_idx: int):
        return any(self._collision(indel, swap_idx) for indel in indels)
    
    def step_del_idx(self, swap_idx: int):
        # Move the indel's position (self._ins_idx) to swap_idx.
        # Move self._del_idx one step in the same direction.
        assert swap_idx != self.ins_idx
        self._del_idx += 1 if swap_idx > self.ins_idx else -1
    
    def _step(self, swap_idx: int):
        self.step_del_idx(swap_idx)
        self._ins_idx = swap_idx
    
    @staticmethod
    def _consistent_rels(curr_rel: int, swap_rel: int):
        if curr_rel & swap_rel or (curr_rel & SUB_N_INT
                                   and swap_rel & SUB_N_INT):
            # Relationship between reference and read base (read_code) and
            # relationship between reference and swap base (swap_code)
            # are consistent, meaning either
            # - both match the reference
            # - one matches and the other potentially matches (i.e. low qual)
            # - one is a substitution and the other could be a substitution
            # - both are substitutions (for each code, code & SUB_N_INT == code)
            return curr_rel
        # Otherwise, i.e.g if one base matches and the other is a substitution,
        # then the relationships are not consistent.
        return 0
    
    def _encode_swap(self, *args, **kwargs) -> bool:
        raise NotImplementedError
    
    def _try_swap(self, *args, **kwargs) -> bool:
        raise NotImplementedError
    
    def sweep(self, muts: bytearray, ref: bytes, read: bytes, qual: bytes,
              dels: List[Deletion], inns: List[Insertion], from3to5: bool,
              tunnel: bool):
        # Move the indel as far as possible in either the 5' or 3' direction.
        while self._try_swap(muts, ref, read, qual, dels, inns, from3to5,
                             tunnel):
            # All actions happen in _try_swap, so loop body is empty.
            pass


class Deletion(Indel):
    @property
    def rank(self):
        return self._ins_idx
    
    @classmethod
    def _encode_swap(cls, ref_base: int, swap_base: int,
                     read_base: int, read_qual: int):
        curr_rel = encode_compare(ref_base, read_base, read_qual)
        swap_rel = encode_compare(swap_base, read_base, read_qual)
        return cls._consistent_rels(curr_rel, swap_rel)

    def _swap(self, muts: bytearray, swap_idx: int, relation: int):
        """
        Arguments
        muts (bytearray): mutation vector
        swap_idx (int): the index in the region to which the deletion moves
                        during this swap
        swap_code (int): the relationship (match, sub, etc.) between the
                         base located at swap_idx and the base in the read
        """
        # The base at swap_idx moves to self.ref_idx, so after the swap, the
        # relationship between self.ref_idx and the read base will be swap_code.
        muts[self.ins_idx] = muts[self.ins_idx] | relation
        # The base at self.ref_idx is marked as a deletion (by definition), so
        # mark the position it moves to (swap_idx) as a deletion too.
        muts[swap_idx] = muts[swap_idx] | DELET_INT
        self._step(swap_idx)
    
    def _try_swap(self, muts: bytearray, ref: bytes, read: bytes, qual: bytes,
                  dels: List[Deletion], inns: List[Insertion],
                  from3to5: bool, tunnel: bool) -> bool:
        swap_idx, tunneled_indels = self._peek_out_of_indel(dels, from3to5)
        read_idx = self.del_idx5 if from3to5 else self.del_idx3
        if (0 <= swap_idx < len(ref) and 0 <= read_idx < len(read)
                and (tunnel or not self.tunneled)
                and not self._collisions(inns, swap_idx)):
            relation = self._encode_swap(ref[self.ins_idx], ref[swap_idx],
                                         read[read_idx], qual[read_idx])
            if relation:
                self._swap(muts, swap_idx, relation)
                for indel in tunneled_indels:
                    indel.step_del_idx(swap_idx)
                return True
        return False


class Insertion(Indel):
    @property
    def rank(self):
        return self._del_idx
    
    def stamp(self, muts: bytearray):
        if 0 <= self.del_idx5 < len(muts):
            muts[self.del_idx5] = muts[self.del_idx5] | INS_5_INT
        if 0 <= self.del_idx3 < len(muts):
            muts[self.del_idx3] = muts[self.del_idx3] | INS_3_INT

    @classmethod
    def _encode_swap(cls, ref_base: int, read_base: int,
                     read_qual: int, swap_base: int, swap_qual: int):
        curr_rel = encode_compare(ref_base, read_base, read_qual)
        swap_rel = encode_compare(ref_base, swap_base, swap_qual)
        return cls._consistent_rels(curr_rel, swap_rel)

    def _swap(self, muts: bytearray, ref_idx: int,
              swap_idx: int, relation: int):
        """
        Arguments
        muts (bytearray): mutation vector
        swap_idx (int): the index in the read to which the deletion moves
                        during this swap
        swap_code (int): the relationship (match, sub, etc.) between the
                         base located at swap_idx and the base in the ref
        """
        # The base at ins_idx moves to swap_idx, so after the swap, the
        # relationship between ref_idx and the read base will be relation.
        muts[ref_idx] = muts[ref_idx] | relation
        self._step(swap_idx)
        # Mark the new positions of the insertion.
        self.stamp(muts)

    def _try_swap(self, muts: bytearray, ref: bytes, read: bytes, qual: bytes,
                  dels: List[Deletion], inns: List[Insertion],
                  from3to5: bool, tunnel: bool) -> bool:
        swap_idx, tunneled_indels = self._peek_out_of_indel(inns, from3to5)
        ref_idx = self.del_idx5 if from3to5 else self.del_idx3
        if (0 <= swap_idx < len(read) and 0 <= ref_idx < len(ref)
                and (tunnel or not self.tunneled)
                and not self._collisions(dels, swap_idx)):
            relation = self._encode_swap(ref[ref_idx],
                                         read[self.ins_idx],
                                         qual[self.ins_idx],
                                         read[swap_idx],
                                         qual[swap_idx])
            if relation:
                self._swap(muts, ref_idx, swap_idx, relation)
                for indel in tunneled_indels:
                    indel.step_del_idx(swap_idx)
                return True
        return False


def sweep_indels(muts: bytearray, ref: bytes, read: bytes, qual: bytes,
                 dels: List[Deletion], inns: List[Insertion], from3to5: bool,
                 tunnel: bool):
    indels = dels + inns
    for indel in indels:
        indel.reset()
    sort_rev = from3to5 != tunnel
    indels.sort(key=lambda indel: indel.rank, reverse=sort_rev)
    while indels:
        indel = indels.pop()
        indel.sweep(muts, ref, read, qual, dels, inns, from3to5, tunnel)
        i = len(indels)
        if sort_rev:
            while i > 0 and indel.rank > indels[i - 1].rank:
                i -= 1
        else:
            while i > 0 and indel.rank < indels[i - 1].rank:
                i -= 1
        if i < len(indels):
            indels.insert(i, indel)


def allindel(muts: bytearray, ref: bytes, read: bytes, qual: bytes,
             dels: List[Deletion], inns: List[Insertion]):
    for from3to5 in (False, True):
        sweep_indels(muts, ref, read, qual, dels, inns, from3to5, True)
        if any(d.tunneled for d in dels) or any(i.tunneled for i in inns):
            sweep_indels(muts, ref, read, qual, dels, inns, from3to5, False)


def parse_cigar(cigar_string: bytes):
    if not cigar_string:
        raise ValueError("CIGAR string was empty.")
    length_matched = 0
    for match in CIG_PATTERN.finditer(cigar_string):
        length_bytes, operation = match.groups()
        length_matched += len(length_bytes) + len(operation)
        if (length_int := int(length_bytes)) < 1:
            raise ValueError("length of CIGAR operation must be >= 1")
        yield operation, length_int
    if length_matched != len(cigar_string):
        raise ValueError(f"Invalid CIGAR string: '{cigar_string.decode()}'")


def op_consumes_ref(op: bytes):
    return op != CIG_INS and op != CIG_SCL


def op_consumes_read(op: bytes):
    return op != CIG_DEL


class SamFlag(object):
    __slots__ = ["paired", "proper", "unmap", "munmap", "rev", "mrev",
                 "first", "second", "secondary", "qcfail", "dup", "supp"]

    MAX_FLAG: int = 2**len(__slots__) - 1
    PATTERN = "".join(["{:0<", str(len(__slots__)), "}"])

    def __init__(self, flag: int):
        if not 0 <= flag <= self.MAX_FLAG:
            raise ValueError(f"Invalid flag: '{flag}'")
        (self.paired, self.proper, self.unmap, self.munmap, self.rev,
         self.mrev, self.first, self.second, self.secondary, self.qcfail,
         self.dup, self.supp) = (x == "1" for x in
                                 self.PATTERN.format(bin(flag)[:1:-1]))


class SamRead(object):
    __slots__ = ["qname", "flag", "rname", "pos", "mapq", "cigar",
                 "tlen", "seq", "qual"]
    
    MIN_FIELDS = 11

    def __init__(self, line: bytes):
        fields = line.rstrip().split(b"\t")
        if len(fields) < self.MIN_FIELDS:
            raise ValueError(f"Invalid SAM line:\n{line}")
        self.qname = fields[0]
        self.flag = SamFlag(int(fields[1]))
        self.rname = fields[2]
        self.pos = int(fields[3])
        self.mapq = int(fields[4])
        self.cigar = fields[5]
        #self.rnext = fields[6]
        #self.pnext = int(fields[7])
        #self.tlen = int(fields[8])
        self.seq = fields[9]
        self.qual = fields[10]
        if len(self) != len(self.qual):
            raise ValueError(f"Lengths of seq ({len(self)}) and qual "
                             f"string {len(self.qual)} did not match.")
        
    def __len__(self):
        return len(self.seq)


def vectorize_read(region_seq: bytes, region_first: int, region_last: int,
                   read: SamRead):
    """
    :param region_seq: str, reference sequence (must contain T, not U)
    :param region_first: int, the first coordinate (w.r.t. the reference
        sequence) of the region of interest (1-indexed)
    :param region_last: int, the last coordinate (w.r.t. the reference
        sequence) of the region of interest, inclusive (1-indexed)
    :param read: SamRead, the read for which to compute mutations
    :return:
    """
    region_length = region_last - region_first + 1
    assert region_length == len(region_seq)
    # current position in the read
    # 0-indexed from beginning of read
    read_start_idx = 0
    read_end_idx = 0
    # position at which the current CIGAR operation starts
    # 0-indexed from beginning of region
    op_start_idx = read.pos - region_first
    # position at which the current CIGAR operation ends (0-indexed)
    # does not include the last position of the operation (like Python slicing)
    # 0-indexed from beginning of region
    op_end_idx = op_start_idx
    # Number of bases truncated from the end of the operation.
    truncated = 0
    # Initialize the mutation vector. Pad the beginning with missing bytes
    # if the read starts after the first position in the region.
    muts = bytearray(BLANK * min(op_start_idx, region_length))
    # Record all deletions and insertions.
    dels: List[Deletion] = list()
    inns: List[Insertion] = list()
    # Read the CIGAR string one operation at a time.
    for cigar_op, op_length in parse_cigar(read.cigar):
        if op_consumes_ref(cigar_op):
            # Advance the end of the operation if it consumes the reference.
            op_end_idx += op_length
        if op_consumes_read(cigar_op):
            read_end_idx += op_length
        if op_end_idx > 0 and op_start_idx < region_length:
            # Run this block once the operation has entered the region.
            if op_start_idx < 0:
                # If the current operation starts before the region,
                # then advance it to the beginning of the region.
                if op_consumes_read(cigar_op):
                    read_start_idx -= op_start_idx
                op_length += op_start_idx
                op_start_idx = 0
            if op_end_idx > region_length:
                # If the current operation ends after the region,
                # then truncate it to the end of the region.
                truncated = op_end_idx - region_length
                op_length -= truncated
                if op_consumes_read(cigar_op):
                    read_end_idx -= truncated
                op_end_idx = region_length
            # Perform an action based on the CIGAR operation and its length.
            if cigar_op == CIG_MAT:
                # Condition: read matches reference
                muts.extend(map(encode_match,
                                read.seq[read_start_idx: read_end_idx],
                                read.qual[read_start_idx: read_end_idx]))
            elif cigar_op == CIG_ALN or cigar_op == CIG_SUB:
                # Condition: read has a match or substitution relative to ref
                muts.extend(map(encode_compare,
                                region_seq[len(muts): len(muts) + op_length],
                                read.seq[read_start_idx: read_end_idx],
                                read.qual[read_start_idx: read_end_idx]))
            elif cigar_op == CIG_DEL:
                # Condition: read contains a deletion w.r.t. the reference.
                dels.extend(Deletion(ref_idx, read_start_idx) for ref_idx
                            in range(len(muts), len(muts) + op_length))
                muts.extend(DELET * op_length)
            elif cigar_op == CIG_INS:
                # Condition: read contains an insertion w.r.t. the reference.
                # Position added to insertions is of the base 3' of the insert.
                inns.extend(Insertion(idx, len(muts)) for idx
                            in range(read_start_idx, read_end_idx))
                # Insertions do not consume the reference, so do not add any
                # information to muts yet. That information is added later.
            elif cigar_op == CIG_SCL:
                # Condition: read contains a soft clipping
                pass
            else:
                raise ValueError(
                    f"Invalid CIGAR operation: '{cigar_op.decode()}'")
        # Advance the start positions to the end of the current operation.
        if truncated:
            op_end_idx += truncated
            if op_consumes_read(cigar_op):
                read_end_idx += truncated
            truncated = 0
        op_start_idx = op_end_idx
        read_start_idx = read_end_idx
    # Pad the end of the mutation vector with any non-covered positions.
    muts.extend(BLANK * (region_length - len(muts)))
    assert len(muts) == region_length
    # Ensure the CIGAR string matched the length of the read.
    if read_start_idx != len(read):
        raise ValueError(
            f"CIGAR string '{read.cigar.decode()}' consumed {read_start_idx} "
            f"bases from read, but read is {len(read)} bases long.")
    # Add insertions to muts.
    for ins in inns:
        ins.stamp(muts)
    # Label all positions that are ambiguous due to indels.
    if dels or inns:
        allindel(muts, region_seq, bytes(read.seq), read.qual, dels, inns)
    return muts


def get_consensus_mut(byte1: int, byte2: int):
    return intersect if (intersect := byte1 & byte2) else byte1 | byte2


def vectorize_pair(region_seq: bytes, region_first: int, region_last: int,
                   read1: SamRead, read2: SamRead):
    muts1 = vectorize_read(region_seq, region_first, region_last, read1)
    muts2 = vectorize_read(region_seq, region_first, region_last, read2)
    return bytearray(map(get_consensus_mut, muts1, muts2))


class SamRecord(object):
    __slots__ = ["read1", "read2"]

    def __init__(self, read1: SamRead, read2: Optional[SamRead] = None):
        self.read1 = read1
        if read2 is not None:
            if self.paired:
                assert read2.flag.paired
                assert read1.qname == read2.qname
                assert read1.rname == read2.rname
                assert read1.flag.first and read2.flag.second
                assert read1.flag.rev != read2.flag.rev
            else:
                raise ValueError("read1 is unpaired, but read2 was given")
        self.read2 = read2
    
    @property
    def read_name(self):
        return self.read1.qname.decode()

    @property
    def ref_name(self):
        return self.read1.rname.decode()

    @property
    def paired(self):
        return self.read1.flag.paired

    def vectorize(self, region_seq: bytes, region_start: int, region_end: int):
        if self.read2 is None:
            return vectorize_read(region_seq, region_start, region_end,
                                  self.read1)
        else:
            return vectorize_pair(region_seq, region_start, region_end,
                                  self.read1, self.read2)



import pandas as pd
import numpy as np
pd.options.mode.chained_assignment = None

def __find_base_in_sequence(sequence, base_type):
    return [i for i, base in enumerate(sequence) if base in base_type]

def __index_selected(row, base_index, base_type, base_pairing, RNAstructure_use_DMS, RNAstructure_use_temp):
    index = list(range(len(row['sequence'])))
    if base_index is not None:
        if isinstance(base_index, int):
            index = [base_index]
        if isinstance(base_index, list):
            index = base_index
        if isinstance(base_index, str):
            assert (count:=row['sequence'].count(base_index)) == 1, f"{count} sequences {base_index} found in sequence of sample {row['sample']} construct {row['construct']} section {row['section']} cluster {row['cluster']} (sequence: {row['sequence']})"
            temp_idx = row['sequence'].find(base_index)
            index = list(range(temp_idx, temp_idx+len(base_index)))

    if base_type is not ['A','C','G','T']:
        index = list(set(index) & set(__find_base_in_sequence(row['sequence'], base_type)))
    
    if base_pairing is not None:
        base_pairing_switch = lambda idx, base, pairing: idx if (base=='.' and pairing) or (base in ['(',')'] and not pairing) else None 
        structure = [base_pairing_switch(i,base,base_pairing) for i, base in enumerate(row['structure_selected'])]
        index = list(set(index) & set([i for i in structure if i is not None]))
    return index


def get_df(df, sample=None, construct=None, section=None, cluster=None, min_cov_bases=0, base_index=None, base_type=['A','C','G','T'], base_pairing=None, RNAstructure_use_DMS=False, RNAstructure_use_temp=False, selected_cols=True, **kwargs)->pd.DataFrame:
    """Get a dataframe with filtered data

    Args:
        df (pd.Dataframe): Dataframe to filter
        sample (list, int, str, optional): Filter rows by sample (list of samples or just a sample). Defaults to None.
        construct (list, int, str, optional): Filter rows by construct (list of constructs or just a construct). Defaults to None.
        section (list, int, str, optional): Filter rows by section (list of sections or just a section). Defaults to None.
        cluster (list, int, str, optional): Filter rows by cluster (list of clusters or just a cluster). Defaults to None.
        min_cov_bases (int, optional): Filter rows by a minimum threshold for base coverage. Defaults to 0.
        base_index (list, int, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base index. Can be a unique sequence in the row's sequence, a list of indexes or a single index. Defaults to None.
        base_type (list, str, optional): Filter per-base attributes (mut_rates, sequence, etc) by base type. Defaults to ['A','C','G','T'].
        base_pairing (bool, optional): Filter per-base attributes (mut_rates, sequence, etc) by expected base pairing. See RNAstructure_use_XXX arguments. Defaults to None.
        RNAstructure_use_DMS (bool, optional): Use DMS for the RNAstructure prediction when filtering by base pairing. Defaults to False.
        RNAstructure_use_temp (bool, optional): Use temperature for the RNAstructure prediction when filtering by base pairing. Defaults to False.
        **kwargs: Additional arguments to pass to filter rows by. Ex: flank='flank_1' will keep only rows with flank=flank_1.

    Returns:
        pd.Dataframe: a filtered dataframe according to the given parameters
    """

    df = df.copy()
    assert df.shape[0] > 0, "Empty dataframe"
    
    df['index_selected'] = pd.Series([[]]*df.shape[0], index=df.index)
    df['structure_selected'] = pd.Series([[]]*df.shape[0], index=df.index)
    df['deltaG_selected'] = pd.Series([[]]*df.shape[0], index=df.index)

    # filter mutation profiles
    df = df.loc[df.worst_cov_bases >= min_cov_bases,:]
    for key, value in kwargs.items():
        locals()[key] = value
    mp_attr = ['sample', 'construct', 'section', 'cluster'] + list(kwargs.keys())
    for attr in mp_attr:
        assert attr in df.columns, f"Attribute {attr} not found in dataframe"
        if eval(attr) is not None:
            if (isinstance(eval(attr), list) or isinstance(eval(attr), tuple)) or isinstance(eval(attr), np.ndarray):
                df = df[df[attr].isin(eval(attr))]
            else:
                df = df[df[attr] == eval(attr)]
    
    if len(df) == 0:
        return df

        
    # filter base profiles
    df.loc[:,'structure_selected'] = df['structure'+('_DMS' if RNAstructure_use_DMS else '')+('_T' if RNAstructure_use_temp else '')] 
    df.loc[:,'deltaG_selected'] = df['deltaG'+('_DMS' if RNAstructure_use_DMS else '')+('_T' if RNAstructure_use_temp else '')] 
    if base_index is not None or base_type is not ['A','C','G','T'] or base_pairing is not None:
        df.loc[:,'index_selected'] = df.apply(lambda row: __index_selected(row, base_index, base_type, base_pairing, RNAstructure_use_DMS, RNAstructure_use_temp), axis=1)
        df = df.loc[df.index_selected.apply(lambda x: len(x) > 0),:]
        bp_attr = ['sequence', 'mut_bases', 'info_bases','del_bases','ins_bases','cov_bases','mut_rates'] + \
            [c for c in df.columns.tolist() if (c.startswith('structure') or c.startswith('mod_bases') or c.startswith('poisson'))]
        for idx, row in df.iterrows():
            for attr in bp_attr:
                # don't filter if the attribute is not an iterable
                if not hasattr(row[attr], '__iter__'):
                    continue
                filtered_cell = [row[attr][i] for i in df.at[idx, 'index_selected']]
                if type(row[attr]) == str:
                    df.at[idx, attr] = ''.join(filtered_cell)
                else:
                    df.at[idx, attr] = np.array(filtered_cell)
    else:
        df.loc[:,'index_selected'] = df.apply(lambda row: list(range(len(row['sequence']))), axis=1)

    if len(df) == 0:
        return df
    try:
        df['unique_id'] = df.apply(lambda row: '_'.join([str(row[attr]) for attr in mp_attr if len(set(df[attr])) > 1]), axis=1)
    except:
        if len(df) > 0:
            "Could not create unique_id column, df: \n {}".format(df)
        else:
            pass

    if not selected_cols:
        df = df.drop(columns=['index_selected', 'structure_selected', 'deltaG_selected','unique_id'])
    return df.copy()



from itertools import repeat
from typing import Any, Callable, Dict, Iterable, Iterator, Tuple


def zip_func_args(func: Callable,
                  iter_args: Iterable[Tuple[Any]],
                  iter_kwargs: Iterable[Dict[str, Any]]):
    """
    Zip a function, a list where each item is a tuple of positional arguments,
    and a list where each item is a dict of keyword arguments into one iterable
    that can be passed to starmap.

    Arguments
    ---------
    function: callable
        Function to be called once for each set of args and kwargs.
    
    iter_args: iterable[tuple[any]]
        Iterable of tuples of positional arguments. Each item in iter_args is a
        tuple of all positional arguments for the function.
    
    iter_kwargs: iterable[dict[str, any]]
        Iterable of dicts of keyword arguments. Each item in iter_kwargs is a
        dict of all keyword arguments for the function.

    Returns
    -------
    iterator[callable, tuple[any], dict[str, any]]
        Arguments that can be passed to the function apply
    """
    return zip(repeat(func), iter_args, iter_kwargs)


def unpack_call(func: Callable, args: Tuple[Any], kwargs: Dict[str, Any]):
    """
    Call a function by unpacking a tuple of positional arguments and a dict of
    keyword arguments.

    Arguments
    ---------
    func: callable
        The function to call
    
    args: tuple[any]
        The positional arguments of the function
    
    kwargs: dict[str, any]
        The keyword arguments of the function
    
    Returns
    -------
    any
        The return value of func called with *args and **kwargs
    """
    return func(*args, **kwargs)


def starstarmap(starmapper: Callable[[Callable, Iterator], Iterable],
                func: Callable,
                iter_args: Iterable[Tuple[Any]],
                iter_kwargs: Iterable[Dict[str, Any]]):
    """
    Call a function multiple times, once for each group of positional and
    keyword arguments in iter_args and iter_kwargs, respectively.
    
    Arguments
    ---------
    starmapper: callable
        The engine performing the starmapping, such as itertools.starmap or
        multiprocessing.Pool.starmap
    
    func: callable
        The function called by starmapper with the arguments
    
    iter_args: iterable[tuple[any]]
        Iterable of tuples of positional arguments. Each item in iter_args is a
        tuple of all positional arguments for the function.
    
    iter_kwargs: iterable[dict[str, any]]
        Iterable of dicts of keyword arguments. Each item in iter_kwargs is a
        dict of all keyword arguments for the function.
    
    Returns
    -------
    iterable
        Iterable of return values of the starmapper function
    """
    return starmapper(unpack_call, zip_func_args(func, iter_args, iter_kwargs))



from __future__ import annotations
from functools import cached_property, wraps
from io import BufferedReader
from typing import Optional

from dreem.util.reads import XamBase, BamVectorSelector, SamVectorSorter
from dreem.util.path import BasePath, XamInPath, XamTempPath
from dreem.vector.vector import *


def _requires_open(func: function):
    @wraps(func)
    def wrapper(self: SamViewer, *args, **kwargs):
        if self._sam_file is None:
            raise ValueError(f"Function '{func.__name__}' requires "
                             "the SAM file to have been opened.")
        return func(self, *args, **kwargs)
    return wrapper


def _reset_seek(func: function):
    @wraps(func)
    @_requires_open
    def wrapper(self: SamViewer, *args, **kwargs):
        prev_pos = self._sam_file.tell()
        ret = func(self, *args, **kwargs)
        self._sam_file.seek(prev_pos)
        return ret
    return wrapper


def _range_of_records(func: function):
    @wraps(func)
    @_reset_seek
    def wrapper(self: SamViewer, start: int, stop: int):
        self._sam_file.seek(start)
        records = iter(func(self))
        if stop is None:
            return records
        else:
            while True:
                if self._sam_file.tell() < stop:
                    yield next(records)
                else:
                    assert self._sam_file.tell() == stop
                    break
    return wrapper


class SamViewer(object):
    def __init__(self, base_path: BasePath, xam_path: XamInPath,
                 ref_name: str, first: int, last: int, spanning: bool,
                 owner: bool = True):
        self.base_path = base_path
        self.xam_path = xam_path
        self.ref_name = ref_name
        self.first = first
        self.last = last
        self.spanning = spanning
        self.owner = owner
        self._sam_path: Optional[XamInPath | XamTempPath] = None
        self._sam_file: Optional[BufferedReader] = None
    
    def __enter__(self):
        # Convert the BAM file to a temporary SAM file
        if self.owner:
            if self.spanning:
                selector = None
                xam_path = self.xam_path
            else:
                xam_base = XamBase(self.base_path, self.xam_path)
                xam_index = xam_base.xam_index
                if not xam_index.path.is_file():
                    assert xam_base.create_index().path == xam_index.path
                selector = BamVectorSelector(self.base_path, self.xam_path)
                xam_path = selector.run(self.ref_name, self.first, self.last)
            sorter = SamVectorSorter(self.base_path, xam_path)
            self._sam_path = sorter.run(name=True)
            if selector:
                selector.clean()
        else:
            self._sam_path = self.xam_path
        self._sam_file = open(self.sam_path.path, "rb")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._sam_file.close()
        self._sam_file = None
        if self.owner:
            self.sam_path.path.unlink()
            self._sam_path = None
    
    @property
    def sam_path(self):
        if self._sam_path:
            return self._sam_path
        raise RuntimeError(f"{self} has no open SAM path.")
    
    @_requires_open
    def _seek_beginning(self):
        self._sam_file.seek(0)
    
    @cached_property
    @_reset_seek
    def _rec1_pos(self):
        self._seek_beginning()
        while (line := self._sam_file.readline()).startswith(SAM_HEADER):
            pass
        return self._sam_file.tell() - len(line)
    
    @_requires_open
    def _seek_rec1(self):
        self._sam_file.seek(self._rec1_pos)
        
    @cached_property
    @_reset_seek
    def paired(self):
        self._seek_rec1()
        first_line = self._sam_file.readline()
        return SamRead(first_line).flag.paired if first_line else False
    
    @_range_of_records
    def _get_records_single(self):
        while read := SamRead(self._sam_file.readline()):
            assert not read.flag.paired
            yield SamRecord(read)

    @_range_of_records
    def _get_records_paired_flexible(self):
        prev_read: Optional[SamRead] = None
        while line := self._sam_file.readline():
            read = SamRead(line)
            assert read.flag.paired
            if prev_read:
                # The previous read has not yet been yielded
                if prev_read.qname == read.qname:
                    # The current read is the mate of the previous read
                    if prev_read.flag.first:
                        yield SamRecord(prev_read, read)
                    else:
                        yield SamRecord(read, prev_read)
                    prev_read = None
                else:
                    # The previous read is paired, but its mate is not in
                    # the SAM file. This situation can happen when only the
                    # reads mapping to a specific region are exported to a
                    # SAM file and only one mate overlaps that region.
                    yield SamRecord(prev_read)
                    # Save the current read so that if its mate is the next
                    # read, it will be returned as a pair.
                    prev_read = read
            else:
                # Save the current read so that if its mate is the next
                # read, it will be returned as a pair.
                prev_read = read
        if prev_read:
            # In case the last record
            yield SamRecord(prev_read)
    
    @_range_of_records
    def _get_records_paired_strict(self):
        while line := self._sam_file.readline():
            yield SamRecord(SamRead(line), SamRead(self._sam_file.readline()))
    
    def get_records(self, start: int, stop: int):
        if self.paired:
            if self.spanning:
                records = self._get_records_paired_strict
            else:
                records = self._get_records_paired_flexible
        else:
            records = self._get_records_single
        return records(start, stop)
    
    @_reset_seek
    def get_batch_indexes(self, batch_size: int):
        if batch_size <= 0:
            raise ValueError("batch_size must be a positive integer")
        n_skip = (self.paired + 1) * (batch_size - 1)
        self._seek_rec1()
        while True:
            yield self._sam_file.tell()
            try:
                line = next(self._sam_file)
            except StopIteration:
                break
            else:
                for _, line in zip(range(n_skip), self._sam_file):
                    pass
                if self.paired:
                    # Compare the current and next query names.
                    try:
                        qname = line.split()[0]
                    except IndexError:
                        pass
                    else:
                        line_next = self._sam_file.readline()
                        try:
                            qname_next = line_next.split()[0]
                        except IndexError:
                            pass
                        else:
                            if qname != qname_next:
                                # If the current and next query names differ
                                # (the lines do not come from two paired mates),
                                # then backtrack to the beginning of line_next.
                                self._sam_file.seek(-len(line_next), 1)



import click
from main import run
from click_option_group import optgroup
from dreem.util.cli import *

@click.command()

@optgroup.group('I/O')
@opti_library
@fastq1
@fastq2
@opto_out_dir

@optgroup.group('Demultiplexing')
@barcode_start
@barcode_length
@max_barcode_mismatches

@optgroup.group('Selection')
@coords
@primers
@fill
@interleaved

@optgroup.group('Miscellaneous')
@verbose

def cli(**args):
    run(**args)

if __name__ == '__main__':
    cli()


"""
Dreemium Messages of Success

Print messages to indicate successful completion of one of Dreemium's programs.
"""

import random


def dms(func):
    def wrapper(*args, **kwargs):
        ret = func(*args, **kwargs)
        print(MessagePicker().quote)
        return ret
    return wrapper


class MessagePicker(object):
    def __init__(self):
        self.num = random.randint(1, len(self._messages) - 2)

    @property
    def message(self):
        return self._messages[self.num].strip()

    @property
    def quote(self):
        return f"Dreemium Message of Success #{self.num}: '{self.message}'"

    _messages = """
    Don't Mess with Silvi
    Detect Mutations Speedily
    Deconvolute Multiple Structures
    Don't Modify with SHAPE
    Dancing at Midnight to Salsa
    Asp-Met-Ser
    Don't Mention "Sunday-funday"
    Duong, Mandy is a Saint
    De-Multiplex, Scott!
    Dreams, Marianne (by Storr)
    Devoted Mentor Silvi
    Doing Minipreps on a Saturday
    Debugging My Software
    Dreaming of a Manuscript in Science
    Drosophila Melanogaster Sevenless
    Doping Makes Semiconductors
    Deprived of Much Sleep
    Drop Mentos into Soda
    Department of Materials Science
    Dijon Mustard Smacks
    DMSO: Moistureless Solvent
    Destroy Microbes with Sulfonamides
    Defected MIT Student
    Drug Many Stem-loops
    Duststorms Move Soil
    Dark Monolith in Space
    Daytime Megawatts from Solar
    Despised Murder-hornets Sting
    Delusional Misidentification Syndrome
    Dry-ice Melts not: Sublimates
    Diatoms Make Shells
    Driving in Massachusetts Sucks
    Dissertation-Making is Stressful
    Deep Marine Sea-vents
    Dielectric Materials Shield
    Do Meticulous Science
    Deceit Makes Secrets
    DNA Multiplexed Sequencing
    Dating Millions of Samples
    Deduction Master Sherlock
    Document Management System
    Degrees° Minutes' Seconds"
    Details Matter Significantly
    Diligence is Mandatory for Success
    Decapitated Moving-trucks got Storrowed
    Dreading a Major Scoop
    Dextrose: Munch Sweetly
    Dying of Med-student Syndrome
    Diagnostic Medical Sonography
    DiMethyl Sulfide :O
    Delivering Mail by Snail
    Dairy MilkShake
    Delaware Makes Startups
    Devoured by Mackerel Sharks
    Dumping Mountains of Snow
    Don't Miss the Sunset
    Drenching Monsoon Season
    David: a Michelangelo Sculpture
    Delicious Midnight Snacking
    Distrust Most Structures
    "Dr. Matty" (Someday)
    Drowsily, Melatonin Slumbers
    Driver Monitoring System
    Disco Music: will Survive
    Driving on Martian Soil
    Doubting My Simulations
    Dashi Miso Soup
    Detect Methionine with Sulfur-35
    Division of Medical Sciences
    Dinner of Meatballs and Spaghetti
    Dear Mansplainer: Silence!
    Degree: Master of Science
    Darn Mangled Sweaters
    A/G/U A/C C/G
    Determine MHV Structure
    Don Masks Safely
    🎼 Do-Mi-So 🎶
    Dynamic Message Sign
    Dinosaurs: Mesozoic Sauropods
    Don't Make up Sh*t
    Daily Mismatched Socks
    "Dülk" Miffs Sarah
    Disrupt Membranes with SDS
    Direct Message me on Slack
    Data from Money is all of Science
    Drive Me to Singapore
    Disrupting Markets with Software
    Double My Stipend (please)
    Duly, Mimes are Silent
    Do Mitochondria Swim?
    Darwin Memorized Species
    Deus ex Machina in Sophocles
    Duckling Matures into Swan
    Downstairs, Monsters Slumber
    Distribute Modules with Setuptools
    Diogenes Mocked Society
    DeRisi Mentored Silvi
    Defuse Mines Safely
    Davy's Minedamp Safety-lamp
    Don't Mash Spiders
    Dragon Marauder Smaug
    Doge Meme $
    Drooling Mouthfuls of Saliva
    Dodos from Mare aux Songes
    Dido Married Sychaeus
    Diseased Martians Succumbed
    Don't Monger Stereotypes
    DaVinci Made Sketches
    Don't Migraines Suck?
    Darth Maul Survived
    Decorated Michelangelo, the Sistine
    Duet of Matter and Spacetime
    DREEM Me up, Scotty!
    Diagnose Methane using Sulfhydryls
    Darkened Marshmallows Smolder
    Develop Machine-learning Solutions
    Daily Meditation for the Soul
    Desmond's Melodious Saxophone
    Daring Motorcycle Stunts
    Demeter Moves the Seasons
    Desert in Mexico: Sonora
    DuPont Manufacturers Sorona
    Diphenhydramine Makes you Sleepy
    Decompose Matrices Spectrally
    Deeds Manifest Sentiments
    Dawn -> Midday -> Sunset
    Demise of Mozart: Streptococci?
    D Minor Sonata
    Duplicating Manuscripts via Selenium
    Deep Mariana Submersible
    Dumbledore Meliorated Snape
    Doomscrolling Magnifies Stress
    Digital Microprocessors: Sentient?
    Doomsday Machine in Strangelove
    Dead Man's Switch
    """.split("\n")



import click
import pandas as pd
from main import run
from dreem.util.cli import *

@click.command()

@optgroup.group('I/O')
@fasta 
@input_dir
@opto_out_dir
@opti_library

@optgroup.group('Selection')
@coords
@primers
@fill

@optgroup.group('Clustering')
@max_clusters
@min_iter
@signal_thresh
@info_thresh
@include_g_u
@include_del
@min_reads
@convergence_cutoff
@num_runs
@n_cpus

@optgroup.group('Miscellaneous')
@verbose


def cli(**args):
    run(**args)

if __name__ == '__main__':
    run()



from EMclustering import EMclustering
from bitvector import BitVector

import numpy as np
import multiprocessing

class ClusteringAnalysis:
    """Launches the clustering algorithm many times to iterate over K_max the number of clusters and N_runs the number of runs.
    
    Parameters from args:
    ---------------------
        
    bitvector: object
        bitvector object. Conatins the processed bitvector and the count of reads per bitvector.
    
    K_max: int
        Number of clusters. The algorithm will iterate from 1 to K_max.
        
    N_runs: int
        Number of runs. The algorithm will be launched N_runs times.
        
    Key methods:
    ------------
    
    run: Run the clustering algorithm many times to iterate over K_max the number of clusters and N_runs the number of runs.
    plot: Plot the results of the clustering analysis. #TODO
        
    """
    
    def __init__(self, bitvector:BitVector, max_clusters:int, N_runs:int, clustering_args):
        self.bitvector = bitvector
        self.K_max = max_clusters
        self.N_runs = N_runs
        self.clustering_args = clustering_args
        # TODO add all

    def run(self):
        '''Run the clustering algorithm many times to iterate over K_max the number of clusters and N_runs the number of runs.
        
        Output:
        -------

        The output is a dictionary containing the results of EMclustering for each K and each run.
        The first level of the dictionary is the number of clusters.
        The second level of the dictionary is a list containing the results of EMclustering for each run.
        The second levels of the dictionary are sorted by performance (log-likelihood).
        
        Example:
        --------
        
        results = {
            K1: [EMclustering1, EMclustering2, ...],
            K2: [EMclustering1, EMclustering2, ...],
            ...
        }
        
        The EMclustering output dictionary contains the following attributes:
            - mu: array (K x D)
                Mean of each cluster.
            - pi: array (K)
                Probability of each cluster.
            - log_likelihood: float
                Log-likelihood of the model.
        
        '''
        # global dT # !! For testing !!
        em = EMclustering(self.bitvector.bv, 1, self.bitvector.read_hist, self.bitvector.base_to_keep, self.bitvector.sequence,
                                **self.clustering_args)
        results = {'K1': [em.run()] }
        for k in range(2,self.K_max+1):
            em = EMclustering(self.bitvector.bv, k, self.bitvector.read_hist, self.bitvector.base_to_keep, self.bitvector.sequence,
                                **self.clustering_args)

            pool = multiprocessing.Pool(processes=self.clustering_args["n_cpus"])
            results['K'+str(k)] = sorted(pool.starmap(em.run, [() for _ in range(self.N_runs)]), key=lambda res: res['log_likelihood'], reverse=True)
            pool.close()
            pool.join()
        
        return results
        

if __name__ == '__main__':
    if False:

        import matplotlib.pyplot as plt # !! For testing !!
        import time, copy # !! For testing !!
        import tracemalloc # !! For testing !!

        # dummy class just for testing
        class testBV_class:
            def __init__(self, bit_vector, read_hist) -> None:
                self.bv = bit_Vector
                self.read_hist = read_hist
                

        exp_path = "/Users/Alberic/Desktop/Pro/RouskinLab/projects/DREEM/"
        bit_Vector_total = np.load(exp_path+"bit_vector.npy")
        read_hist_total =  np.load(exp_path+"read_hist.npy")

        with open(exp_path+"data_EM_analysis_v2.txt", 'a') as f:
            f.write("N_reads n_cpu dT mean_memory peak_memory \n")
            f.close()
        
        for N_partial in np.geomspace(10000, 120000, 5).astype(np.int64):

            N_partial = min(N_partial, bit_Vector_total.shape[0])

            for n_cpu in range(1, 11):

                print("Starting experiment with {} reads and {} cpus".format(N_partial, n_cpu))
                
                tracemalloc.start()

                # Get part of the total reads
                bit_Vector = copy.copy(bit_Vector_total[:N_partial])
                read_hist = copy.copy(read_hist_total[:N_partial])
                BV_class = testBV_class(bit_vector=bit_Vector, read_hist=read_hist)
                
                # Run EM and record stats
                clustering_args = dict(
                    min_iter = 10,
                    signal_thresh = 0.5, 
                    info_thresh = 0.5, 
                    include_g_u = True, 
                    include_del = True, 
                    min_reads = 10,
                    convergence_cutoff = 0.5,
                    num_runs = 100,
                    n_cpus = n_cpu,
                    verbose = True
                )

                clustering = ClusteringAnalysis(BV_class, 2, 10, clustering_args)
                clustering.run()
                
                # Log experiment stats
                memory_stats = [ mem/1e6 for mem in tracemalloc.get_traced_memory() ]
                
                print("Mean and max memory [MB]:", memory_stats)
                print("Total time [s]", "{:.1f}s".format(dT)) # !! For testing !!
                
                log_data = np.array((N_partial,
                                        n_cpu,
                                        dT,
                                        memory_stats[0], 
                                        memory_stats[1] ))
    
                with open(exp_path+"data_EM_analysis_v2.txt", 'a') as f:
                    np.savetxt(f, log_data, newline=" ")
                    f.write("\n")
                    f.close()

                # Stop memory tracing and clean memory
                tracemalloc.stop()
                bit_Vector = None
                read_hist = None
                time.sleep(5)
                
                # Comparing output with previous software -> different test
                # mu_reference = np.load("/Users/Alberic/Desktop/Pro/RouskinLab/projects/DREEM/result.npy")
                # print("Reference matched:",(mu_reference == result["mu"]).all())



import click
from main import run
from click_option_group import optgroup
from dreem.util.cli import *

@click.command()

@argi_fasta
@opti_fastqu
@opti_fastqi
@opti_fastq1
@opti_fastq2
@opto_out_dir

def cli(*args, **kwargs):
    run(*args, **kwargs)

if __name__ == '__main__':
    cli()



import dreem
import os, sys
import click
import pandas as pd
import dreem.util as util
import numpy as np
import string
import json
import jsbeautifier

from dreem.aggregate.library_samples import get_samples_info, get_library_info
from dreem.aggregate.rnastructure import add_rnastructure_predictions
from dreem.aggregate.poisson import compute_conf_interval
from dreem.util.cli import INPUT_DIR, LIBRARY, SAMPLES, SAMPLE, CLUSTERING_FILE, OUT_DIR, FASTA, RNASTRUCTURE_PATH, RNASTRUCTURE_TEMPERATURE, RNASTRUCTURE_FOLD_ARGS, RNASTRUCTURE_DMS, RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE, RNASTRUCTURE_DMS_MAX_PAIRED_VALUE, POISSON, VERBOSE, COORDS, PRIMERS, FILL, RNASTRUCTURE_PARTITION, RNASTRUCTURE_PROBABILITY
sys.path.append(os.path.dirname(__file__))
from mutation_count import generate_mut_profile_from_bit_vector
from dreem.util.files_sanity import check_library, check_samples
from dreem.aggregate.rnastructure import RNAstructure
from dreem.util.seq import parse_fasta
from dreem.util.dump import *
import logging

def run(bv_files:list, library:str=LIBRARY, samples:str=SAMPLES, sample:str=SAMPLE, clustering_file:str=CLUSTERING_FILE, out_dir:str=OUT_DIR, fasta:str = FASTA, rnastructure_path:str=RNASTRUCTURE_PATH, rnastructure_temperature:bool=RNASTRUCTURE_TEMPERATURE, rnastructure_fold_args:str=RNASTRUCTURE_FOLD_ARGS, rnastructure_dms:bool=RNASTRUCTURE_DMS, rnastructure_dms_min_unpaired_value:int=RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE, rnastructure_dms_max_paired_value:int=RNASTRUCTURE_DMS_MAX_PAIRED_VALUE, rnastructure_partition:bool=RNASTRUCTURE_PARTITION, rnastructure_probability:bool=RNASTRUCTURE_PROBABILITY, poisson:bool=POISSON, verbose:bool=VERBOSE, coords:str=COORDS, primers:str=PRIMERS, fill:bool=FILL):
    """Run the aggregate module.

    Reads in the bit vector files and aggregates them into a single file named [output]/output/aggregate/[name].csv.
    Adds on the library information, the sample information, RNAstructure predictions, and Poisson confidence intervals.

    Parameters from args:
    ---------------------

    bv_files: str
        Path to the bit vector files folders.
    library: str
        Path to a csv file with the library information.
    samples: str
        Path to a csv file with the sample information.
    sample: str
        Name to identify the row in samples.csv. Also the name for the output file. Default is the containing folder name.
    clustering_file: str
        Path to the clustering.json file.
    out_dir: str
        Path to the output folder (the sample).
    fasta: str
        Path to the fasta file.
    rnastructure_path: str
        Path to RNAstructure, to predict structure and free energy.
    rnastructure_temperature: bool
        Use sample.csv temperature values for RNAstructure.
    rnastructure_fold_args: str
        Arguments to pass to RNAstructure fold.
    rnastructure_dms: bool
        Use the DMS signal to amke predictions with RNAstructure.
    rnastructure_dms_min_unpaired_value: int
        Minimum unpaired value for using the dms signal as an input for RNAstructure.
    rnastructure_dms_max_paired_value: int
        Maximum paired value for using the dms signal as an input for RNAstructure.
    rnastructure_partition: bool
        Use RNAstructure partition function to predict free energy.
    rnastructure_probability: bool
        Use RNAstructure probability to predict free energy.
    poisson: bool
        Predict Poisson confidence intervals.
    verbose: bool
        Verbose output.
    coords: tuple
        coordinates for reference: '-c ref-name first last'
    primers: tuple
        primers for reference: '-p ref-name fwd rev'
    fill: bool
        Fill in coordinates of reference sequences for which neither coordinates nor primers were given (default: no).
        
    Returns:
    --------
    1 if successful, 0 otherwise.

    """

    # Extract the arguments
    library = check_library(pd.read_csv(library), fasta, out_dir) if library is not None else None
    library['section_boundaries'] = library.apply(lambda x: str(x['section_start']) + '-' + str(x['section_end']), axis=1)
    df_samples = check_samples(pd.read_csv(samples)) if samples is not None else None
    fasta = pd.DataFrame({k.decode("utf-8") :v.decode("utf-8")  for k,v in parse_fasta(fasta)}, index=[0]).T.reset_index().rename(columns={"index":"construct", 0:"sequence"})

    rnastructure = {
        'path': rnastructure_path,
        'temperature': rnastructure_temperature,
        'fold_args': rnastructure_fold_args,
        'dms': rnastructure_dms,
        'dms_min_unpaired_value': rnastructure_dms_min_unpaired_value,
        'dms_max_paired_value': rnastructure_dms_max_paired_value,
        'partition': rnastructure_partition,
        'probability': rnastructure_probability,
        'temp_folder': os.path.join(out_dir, 'temp', 'rnastructure')
    }
    
    # Find a name for the sample
    if sample is None:
        if df_samples is not None:
            raise ValueError('If samples is specified, sample must also be specified.')
        sample = os.path.basename(os.path.normpath(out_dir))
    
    # Make folders
    os.makedirs(out_dir, exist_ok=True)
    os.makedirs(os.path.join(out_dir, 'temp'), exist_ok=True)

    # Read in the bit vectors
    if clustering_file is not None:
        with open(clustering_file, 'r') as f:
            clustering_file = json.load(f)    
    
    mut_profiles = {}
    print('Reading in bit vectors from {}...'.format(bv_files))
    for bv in bv_files:
        construct, section = bv.split('/')[-2], bv.split('/')[-1].split('.')[0]
        
        assert len(library[(library['construct'] == construct)&(library['section_boundaries'] == section)]) < 2, 'Library information not unique for construct {} section {}'.format(construct, section)
        assert len(library[(library['construct'] == construct)&(library['section_boundaries'] == section)]) > 0, 'Library information not existing for construct {} section {}'.format(construct, section)
        section = library[(library['construct'] == construct)&(library['section_boundaries'] == section)]['section'].values[0]
        
        if not len(os.listdir(bv)) > 0:
            logging.warning('No bit vectors found for construct {}'.format(construct))
            continue
        
        if construct not in mut_profiles:
            mut_profiles[construct] = {'sequence': fasta[fasta['construct'] == construct]['sequence'].values[0]}
        # Add the library information
        mut_profiles[construct] = {**get_library_info(library, construct, verbose=verbose), **mut_profiles[construct]}

        assert library[(library['construct'] == construct)&(library['section'] == section)].shape[0] == 1, 'Library information not found for construct {} section {}'.format(construct, section)
        mut_profiles[construct][section] = {}
        mut_profiles[construct][section]['section_start'] = library[(library['construct'] == construct)&(library['section'] == section)]['section_start'].values[0]
        mut_profiles[construct][section]['section_end'] = library[(library['construct'] == construct)&(library['section'] == section)]['section_end'].values[0]
        mut_profiles[construct][section]['pop_avg'] = generate_mut_profile_from_bit_vector(bv, clustering_file=clustering_file, verbose=verbose)
        mut_profiles[construct][section]['sequence'] = mut_profiles[construct][section]['pop_avg'].pop('sequence')
        assert mut_profiles[construct]['sequence'][mut_profiles[construct][section]['section_start']-1:mut_profiles[construct][section]['section_end']] == mut_profiles[construct][section]['sequence'], 'Sequence mismatch for construct {} section {}: {} vs {}'.format(construct, section, mut_profiles[construct]['sequence'][mut_profiles[construct][section]['section_start']-1:mut_profiles[construct][section]['section_end']], mut_profiles[construct][section]['sequence'])
        for col in ['num_aligned']:
            mut_profiles[construct][col] = mut_profiles[construct][section]['pop_avg'].pop(col)

    print('Done.')
    if df_samples is not None:
        # Add the sample information
        print('Adding sample information...')
        mut_profiles = {**mut_profiles, **get_samples_info(df_samples, sample, verbose=verbose)}
        print('Done.')
        
    if rnastructure['path'] is not None:
        rna = RNAstructure(rnastructure)
    
    print('Computing confidence intervals and RNAstructure predictions...')
    for construct in mut_profiles:
        if type(mut_profiles[construct]) is not dict:
            continue

        for section in mut_profiles[construct]:
            if type(mut_profiles[construct][section]) is not dict:
                continue
            # Add RNAstructure predictions

            if rnastructure['path'] is not None:
                mh = rna.run(mut_profiles[construct][section], sample, sequence_only=True)
                mut_profiles[construct][section] = {**mut_profiles[construct][section], **mh}

            if poisson:
                for cluster in mut_profiles[construct][section]:
                    # Add Poisson confidence intervals
                    if type(mut_profiles[construct][section][cluster]) is not dict:
                        continue
                    d = mut_profiles[construct][section][cluster]
                    mut_profiles[construct][section][cluster] = {**d, **compute_conf_interval(info_bases=d['info_bases'], mut_bases=d['mut_bases'])}
    print('Done.')
    # Write the output
    print('Cast dictionary, size:', sys.getsizeof(mut_profiles))
    out = cast_dict(mut_profiles)
    print('Done.')
    print('Sort dictionary, size:', sys.getsizeof(mut_profiles))
    out = sort_dict(out)
    print('Done.')
    options = jsbeautifier.default_options()
    options.indent_size = 4
    print('Dump the json, size', sys.getsizeof(json.dumps(out, cls=NpEncoder)))
    with open(os.path.join(out_dir, sample + '.json'), 'w') as f:
        f.write(json.dumps(out, cls=NpEncoder))
    print('Done.')
    print('Done aggregating the data.')
    return 1





import dreem.util as util
import os  
import pandas as pd
import numpy as np
from scipy import signal
import datetime
from dreem.util.cli import FASTQ1, FASTQ2, LIBRARY, OUT_DIR, MAX_BARCODE_MISMATCHES, VERBOSE, DEFAULT_INTERLEAVED_INPUT, COORDS, PRIMERS, FILL, FASTA
from dreem.util.files_sanity import check_library

def demultiplex(f1: str = FASTQ1, f2: str = FASTQ2, fasta: str = FASTA, interleaved: bool = DEFAULT_INTERLEAVED_INPUT, library: str = LIBRARY, output_folder: str = OUT_DIR, max_barcode_mismatches: int = MAX_BARCODE_MISMATCHES, verbose: bool = VERBOSE):
    """Demultiplex a pair of FASTQ files.

    Publishes to `output_folder` a pair of FASTQ files for each construct, named {construct}_R1.fastq and {construct}_R2.fastq.

    Parameters
    ----------
    f1: str
        Path to the FASTQ file, forward primer.
    f2: str
        Path to the FASTQ file, reverse primer.
    fasta: str
        Path to the FASTA file containing the primers.
    interleaved: bool
        Whether the FASTQ files are interleaved.
    library: pd.DataFrame
        Columns are (non-exclusively): ['construct', 'barcode_start', 'barcode']
    output_folder: str
        Where to output the results.
    max_barcode_mismatches: int
        Maximum number of mutations allowed on the barcode.
    verbose: bool
        Whether to print the progress.
        
    returns
    -------
    1 if successful, 0 otherwise.
    """
    
    library = check_library(pd.read_csv(library), fasta)
    constructs = library['construct'].unique()
    barcodes = library['barcode'].unique()

    assert len(constructs) == len(barcodes)
    for construct, barcode in zip(constructs, barcodes):
        # check if the barcode and the construct are on the same row
        assert library.loc[library['construct']==construct, 'barcode'].values[0] == barcode

    if not os.path.exists(output_folder):
        os.makedirs(output_folder)        

    # copy the reads from the fastq files that contain the barcode in a fastq file named after the construct in the output folder
    primers = [1,2] if not interleaved else [1]
    for primer in primers:
        
        fq = f1 if primer == 1 else f2
        
        # infos for the report
        perfect_matches_count = 0
        off_matches_count = {k:0 for k in range(1, max_barcode_mismatches+1)}
        lost_reads_count = 0
        count_per_construct = {construct:0 for construct in constructs}
        barcode_shifts = []  
        
        construct_fastq_was_written = {construct:False for construct in constructs}
                        
        with open(fq, 'r') as f:
            while True:
                header, sequence, quality = read_fastq_line(f)
                if not header:
                    break
                
                flag_match = False
                for construct, barcode in zip(constructs, barcodes):
                    
                    minimal_corr_score = worst_matching_score(barcode, max_barcode_mismatches)
                    
                    if primer == 2:
                        barcode = reverse_complement(barcode)
                    
                    corr = compute_correlation(embed_sequence_as_binary(barcode), embed_sequence_as_binary(sequence))
                    is_match = barcode_in_read(corr, minimal_corr_score)
                    if is_match:
                        if max(corr) == 1:
                            perfect_matches_count += 1
                        else:
                            for k in range(1, max_barcode_mismatches+1):
                                if max(corr) >= worst_matching_score(barcode, k):
                                    off_matches_count[k] += 1
                                    break
                            
                        barcode_start = library.loc[library['construct']==construct, 'barcode_start'].values[0]
                        if primer == 2:
                            barcode_shifts.append(len(sequence) - np.argmax(corr) - barcode_start - len(barcode))
                        else:
                            barcode_shifts.append(np.argmax(corr) - barcode_start)
                            
                        count_per_construct[construct] += 1
                        
                        if not construct_fastq_was_written[construct]:
                            with open(os.path.join(output_folder, construct + '_R' + str(primer) + '.fastq'), 'w') as g:
                                write_fastq_line(g, header, sequence, quality)
                            construct_fastq_was_written[construct] = True
                        else:
                            with open(os.path.join(output_folder, construct + '_R' + str(primer) + '.fastq'), 'a') as g:
                                write_fastq_line(g, header, sequence, quality)
                            
                        flag_match = True
                        break
                    
                if not flag_match:
                    lost_reads_count += 1
                    with open(os.path.join(output_folder, 'lost_reads_R' + str(primer) + '.fastq'), 'a') as g:
                        write_fastq_line(g, header, sequence, quality)
                        
        write_report(fq, output_folder, perfect_matches_count, off_matches_count, lost_reads_count, barcode_shifts, count_per_construct)
    return 1

def write_report(fastq, output_folder, perfect_matches_count, off_matches_count, lost_reads_count, barcode_shifts, count_per_construct):
    """Write a report of the demultiplexing process for the given fastq file."""
    with open(os.path.join(output_folder, 'report.txt'), 'a') as f:
        f.write("Time: " + str(datetime.datetime.now()) + "\n")
        f.write('\n'+'='*len('Demultiplexing report for ' + fastq) + '\n')
        f.write('Demultiplexing report for ' + fastq + '\n')
        f.write('='*len('Demultiplexing report for ' + fastq) + '\n')
        f.write('Count of perfect matches: ' + str(perfect_matches_count) + '\n')
        for k in off_matches_count:
            f.write('Count of ' + str(k) + '-off matches: ' + str(off_matches_count[k]) + '\n')
        f.write('Count of lost reads: ' + str(lost_reads_count) + '\n')
        f.write('Count of reads per barcode position: ' + str(bin_positions(barcode_shifts)) + '\n')
        f.write('\nCount of reads per construct: ' + '\n' + '-'*len('Count of reads per construct:') + '\n')
        for construct in count_per_construct:
            f.write(construct + ': ' + str(count_per_construct[construct]) + '\n')
        f.write('='*len('Demultiplexing report for ' + fastq) + '\n')
        
def worst_matching_score(barcode, max_muts=1):
    return 1. - float(max_muts)/len(barcode) -1E-9
                                    
def bin_positions(positions):
    """Turns a list of positions into a dictionary of bins."""
    bins = {}
    for pos in positions:
        if pos not in bins:
            bins[pos] = 0
        bins[pos] += 1
    bins = {k: bins[k] for k in sorted(bins)}
    return bins

def read_fastq_line(f):
    """Read the line of a fastq file and return a tuple (header, sequence, quality)"""
    header = f.readline().strip()
    sequence = f.readline().strip()
    f.readline()
    quality = f.readline().strip()
    return header, sequence, quality

def write_fastq_line(f, header, sequence, quality):
    """Write a fastq file line."""
    f.write(header + '\n')
    f.write(sequence + '\n')
    f.write('+\n')
    f.write(quality + '\n')

def embed_sequence_as_binary(sequence):
    """Each sequence is represented as 4 binary vector of length len(sequence), one per base A C T G."""
    return np.array([np.array([base == 'A', base == 'C', base == 'T', base == 'G']) for base in sequence], dtype=np.int8).T

def compute_correlation(barcode, read):
    """ Use the correlation between the barcode and the read to determine if the barcode is in the read."""
    return signal.correlate(read,barcode, mode="valid", method="auto").squeeze()/barcode.shape[1]

def barcode_in_read(corr, min_corr_score):
    """Return True if the correlation score is above the threshold, False otherwise."""
    return max(corr) > min_corr_score

def hamming_distance(s1, s2):
    assert len(s1) == len(s2)
    return sum(ch1 != ch2 for ch1, ch2 in zip(s1, s2))

def reverse_complement(seq):
    return seq[::-1].translate(str.maketrans('ATCG', 'TAGC'))

def next_base(base):
    return {'A':'T','T':'C','C':'G','G':'A',0:1}[base]

def run(fastq1:str = FASTQ1, fastq2:str = FASTQ2, fasta:str = FASTA, interleaved:bool=DEFAULT_INTERLEAVED_INPUT, library:str = LIBRARY, out_dir:str = OUT_DIR, max_barcode_mismatches:str = MAX_BARCODE_MISMATCHES, verbose:bool = VERBOSE):
    """Run the demultiplexing pipeline.

    Demultiplexes the reads and outputs one fastq file per construct in the directory `output_path`, using `temp_path` as a temp directory.

    Parameters from args:
    -----------------------
    fastq1: str
        Path to the FASTQ file or list of paths to the FASTQ files, forward primer.
    fastq2: str
        Path to the FASTQ file or list of paths to the FASTQ files, reverse primer.
    fasta: str
        Path to the FASTA file.
    interleaved: bool
        If True, the FASTQ files are interleaved.
    library: str
        Path to the library file. Columns are (non-excusively): ['construct', 'barcode_start', 'barcode']
    out_dir: str
        Name of the output directory.
    max_barcode_mismatches: int
        Maximum number of mutations allowed on the barcode.
    verbose: bool
        Print progress to stdout (default: no).
        
    Returns
    -------
    1 if successful, 0 otherwise.

    """

    assert os.path.isfile(library), "Library file not found"

    # Get the paths
    fastq1 = fastq1 if type(fastq1) == list else [fastq1]
    if not interleaved:
        fastq2 = fastq2 if type(fastq2) == list else [fastq2]
        for f2 in fastq2:
            assert os.path.isfile(f2), "FASTQ file not found"
            
    # Make the folders
    os.makedirs(out_dir, exist_ok=True)

    # Remove the report file if it exists
    report_path = os.path.join(out_dir, 'report.txt')
    if os.path.exists(report_path):
        os.remove(report_path)
    
    # Demultiplex
    for f1, f2 in zip(fastq1, fastq2):
        assert demultiplex(f1, f2, fasta, interleaved, library, out_dir, max_barcode_mismatches), "Demultiplexing failed"
    return 1


import yaml, os

def read_sample_attributes():
    return {
        'mandatory': {
            'all': [
                'sample',
                'user',
                'date',
                'exp_env',
                'temperature_k',
                'inc_time_tot_secs',
                'DMS_conc_mM',
            ],
            'in_vitro': [
                'buffer',
            ],
            'in_vivo': [
                'cell_line',
            ],
        },
    }




import pathlib
import sys,os
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))

from dreem.align.align import all_refs, each_ref


def run(out_dir: str, fasta: str,
        fastqu: str = "", fastqi: str = "",
        fastq1: str = "", fastq2: str = "",
        fastqu_dir: str = "", fastqi_dir: str = "",
        fastq12_dir: str = "", **kwargs):
    """Run the alignment module.

    Aligns the reads to the reference genome and outputs one bam file per construct in the directory `output_path`, using `temp_path` as a temp directory.

     /out_dir/
        {sample_1}/
            —| {ref_1}.bam
            —| {ref_2}.bam
            —| ...
        {sample_2}/
        ...

    Parameters from args:
    -----------------------
    out_dir: str
        Path to the output folder (in general the sample). 
    fasta: str
        Path to the reference FASTA file.
    fastq1: str
        Path to the FASTQ file or list of paths to the FASTQ files, forward primer.
    fastq2: str
        Path to the FASTQ file or list of paths to the FASTQ files, reverse primer.
    demultiplexed: bool
        Whether the FASTQ files were demultiplexed (default: False).   
        If True:
            Assume that each FASTQ file contains reads from ONE sample and ONE reference.
            This happens after the (optional) demultiplexing step, whose output follows this structure:
            Every FASTQ file given must be named with the reference name, and if paired-end, followed by the mate (1 or 2).
            Every FASTQ file must be inside a directory with the sample name.
                {sample_1}/
                    |- {ref_1}_R1.fastq
                    |- {ref_1}_R2.fastq
                    |- {ref_2}_R1.fastq
                    |- {ref_2}_R2.fastq
                    |- ...
                {sample_2}/
                ...
            Since each FASTQ is known to correspond to only one reference (and may misalign if aligned to all references in the FASTA file),
            for each reference in the FASTA file, create a new FASTA file with only that reference and align all FASTQ files with the corresponding name to that reference.
        If False (default):
            Assume each FASTQ file contains reads from ONE sample and ONE OR MORE references.
            This is the typical structure of the FASTQ files directly from a sequencer.
            Every FASTQ file given must be named with the sample name, and if paired-end, followed by the mate (1 or 2).
            The directory of the FASTQ files does not matter.
                {sample_1}_R1.fastq
                {sample_1}_R2.fastq
                {sample_2}_R1.fastq
                {sample_2}_R2.fastq
                ...
    interleaved: bool
        Whether the FASTQ files are interleaved (default: False).
    
    """
    demult_args = fastqu_dir, fastqi_dir, fastq12_dir
    non_demult_args = fastqu, fastqi, fastq1, fastq2
    demultiplexed = any(map(bool, (demult_args)))
    non_demultiplexed = any(map(bool, (non_demult_args)))
    if demultiplexed and non_demultiplexed:
        raise ValueError("Both non-demultiplexed FASTQ files "
                         "and demultiplexed FASTQ directories were given.")
    if demultiplexed:
        each_ref(out_dir, fasta, *demult_args, **kwargs)
    elif non_demultiplexed:
        all_refs(out_dir, fasta, *non_demult_args, **kwargs)
    else:
        raise ValueError("No FASTQ input files were given")



import os
from typing import List, Tuple

import pandas as pd
import subprocess
import warnings

from dreem.util.util import DNA, run_cmd
from dreem.util.cli import OUT_DIR, LIBRARY, COORDS, PRIMERS, FILL, PARALLEL
from dreem.util.path import BAM_EXT
from dreem.vector.mprofile import VectorWriterSpawner
from dreem.util.files_sanity import check_library


def add_coords_from_library(library_path: str,
                            coords: List[Tuple[str, int, int]],
                            fasta: str,
                            out_dir: str):
    library = check_library(pd.read_csv(library_path), fasta, out_dir)
    for row in library.index:
        construct = os.path.splitext(str(library.loc[row, "construct"]))[0]
        first = int(library.loc[row, "section_start"])
        last = int(library.loc[row, "section_end"])
        coord = (construct, first, last)
        if coord not in coords:
            coords.append(coord)


def encode_primers(primers: List[Tuple[str, str, str]]):
    return [(ref, DNA(fwd.encode()), DNA(rev.encode()))
            for ref, fwd, rev in primers]


def run(fasta: str, bam_dirs: List[str], out_dir: str = OUT_DIR,
        library: str = LIBRARY, coords: list = COORDS, primers: list = PRIMERS,
        fill: bool = FILL, parallel: str = PARALLEL):
    """
    Run the vectoring step.
    Generate a vector encoding mutations for each read
    (or read pair, if paired-end).

    FASTA (path): reference sequence(s) in FASTA format

    BAM_FILES (paths): list of one or more alignment files (space separated)
                       in BAM format
    """

    # read library
    if library:
        add_coords_from_library(library, coords, fasta, out_dir)
                        
    # Compute mutation vectors for each BAM file
    bam_files = [os.path.join(bam_dir, bam_file)
                 for bam_dir in bam_dirs
                 for bam_file in os.listdir(bam_dir)
                 if bam_file.endswith(BAM_EXT)]
    writers = VectorWriterSpawner(out_dir, fasta, bam_files, coords,
                                  encode_primers(primers), fill, parallel)
    writers.profile()



import os


# Global parameters
NUM_PROCESSES = cpus if (cpus := os.cpu_count()) else 1
BASE_COLORS = {"A": "#D3822A", "C": "#5AB4E5", "G": "#EAE958", "T": "#357766"}
PHRED_ENCODING = 33
BUFFER_LENGTH = 10000




import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
import pandas as pd
import dreem.util as util
from dreem import vectoring
from dreem.test import files_generator
from dreem.test.files_generator import test_files_dir, input_dir, prediction_dir, output_dir
import pytest
from dreem.test.sample_profile import sample_profile, constructs, sections

module = 'vectoring'
sample_name = 'test_set_1'

module_input = os.path.join(input_dir, module)
module_expected = os.path.join(prediction_dir, module)
module_output =  os.path.join(output_dir, module)

inputs = ['bam','fasta','library']
outputs = ['bitvector']

def test_make_files():
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'output', module)))
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'input', module)))
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'expected_output', module)))

    if not os.path.exists(os.path.join(test_files_dir, 'input', module)):
        os.makedirs(os.path.join(test_files_dir, 'input', module))
    files_generator.generate_files(sample_profile, module, inputs, outputs, test_files_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, inputs, input_dir, sample_name)
    files_generator.assert_files_exist(sample_profile, module, outputs, prediction_dir, sample_name)

def test_run():
    for sample in os.listdir(module_input):
        vectoring.run(
            bam_dirs = [os.path.join(module_input, sample, f) for f in os.listdir(os.path.join(module_input, sample)) if f.endswith('.bam')],
            out_dir = os.path.join(module_output),
            fasta = os.path.join(module_input, sample, 'reference.fasta'),
            library = os.path.join(module_input, sample, 'library.csv'),
            )

def test_output_exists():        
    files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, 'vectoring/'+sample_name)

def test_files_are_equal():
    files_generator.assert_files_exist(sample_profile, module, outputs, output_dir, 'vectoring/'+sample_name)
    for sample in os.listdir(module_input):
        for construct in os.listdir(os.path.join(module_expected,sample)):
            for section in os.listdir(os.path.join(module_expected,sample,construct)):
                p = pd.read_orc(os.path.abspath(os.path.join(module_expected,sample,construct,section,'0.orc')))
                o = pd.read_orc(os.path.abspath(os.path.join(module_output,'vectoring',sample,construct,section,'0.orc')))
                for pp, oo in zip(p.columns, o.columns):
                    assert pp == oo, 'Columns are not the same in expected col {} and result col {} for sample {} construct {} section {}'.format(pp, oo, sample, construct, section)
                    assert (p[pp] == o[pp]).all(), 'Values are not the same in expected {} and result {} for sample {} construct {} section {}'.format(pp, oo, sample, construct, section)

if __name__ == '__main__':
    # remove all files
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'output', module)))
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'input', module)))
    os.system('rm -rf {}'.format(os.path.join(test_files_dir, 'expected_output', module)))
    test_make_files()
    test_run()


import yaml
import os, sys
sys.path.append(os.path.dirname(os.path.dirname(__file__)))
from dreem import util
import dreem # import all the macros from the cli.py file
from dreem.util.cli import *


def run(fasta:str, fastq1:str, fastq2:str=FASTQ2, library:str=LIBRARY, samples:str=SAMPLES, out_dir:str=OUT_DIR, 
        demultiplexing:bool=DEFAULT_DEMULTIPLEXED, clustering:bool=CLUSTERING,
        primers:str=PRIMERS, coords:str=COORDS, fill:bool=FILL, parallel:bool=PARALLEL, interleaved:bool=DEFAULT_INTERLEAVED_INPUT,
        barcode_start:int=BARCODE_START, barcode_length:int=BARCODE_LENGTH, max_barcode_mismatches:int=MAX_BARCODE_MISMATCHES,
        max_clusters:int=MAX_CLUSTERS, signal_thresh:float=SIGNAL_THRESH, info_thresh:float=INFO_THRESH, include_g_u:bool=INCLUDE_G_U, include_del:bool=INCLUDE_DEL, min_reads:int=MIN_READS, convergence_cutoff:float=CONVERGENCE_CUTOFF, min_iter:int=MIN_ITER, num_runs:int=NUM_RUNS, n_cpus:int=N_CPUS,
        rnastructure_path:str=RNASTRUCTURE_PATH, rnastructure_temperature:bool=RNASTRUCTURE_TEMPERATURE, rnastructure_fold_args:str=RNASTRUCTURE_FOLD_ARGS, rnastructure_dms:bool=RNASTRUCTURE_DMS, rnastructure_dms_min_unpaired_value:int=RNASTRUCTURE_DMS_MIN_UNPAIRED_VALUE, rnastructure_dms_max_paired_value:int=RNASTRUCTURE_DMS_MAX_PAIRED_VALUE, rnastructure_partition:bool=RNASTRUCTURE_PARTITION, rnastructure_probability:bool=RNASTRUCTURE_PROBABILITY, poisson:bool=POISSON,verbose=VERBOSE):
    """Run DREEM. The input arguments are parsed from the command line. They correspond to the parameters of the functions in the other modules.

    Parameters
    ----------
    out_dir: str
        Path to the output folder.
    fastq1: str
        Path to the fastq file 1.
    fastq2: str
        Path to the fastq file 2.
    fasta: str
        Path to the fasta file.
    library: str
        Path to the library file.
    samples: str
        Path to the samples file.
    
    demultiplexing: bool
        Run demultiplexing.
    clustering: bool
        Run clustering.
        
    primers: str
        coordinates for reference: '-c ref-name first last'
    coords: str
        primers for reference: '-p ref-name fwd rev'
    fill: str
        Fill in coordinates of reference sequences for which neither coordinates nor primers were given (default: no).
    parallel: str
        "Parallelize the processing of mutational PROFILES or READS within each profile, turn parallelization OFF, or AUTO matically choose the parallelization method (default: auto).
    interleaved: bool
        Fastq files are interleaved.
    
    barcode_start: int
        Start position of the barcode.
    barcode_length: int
        Length of the barcode.
    max_barcode_mismatches: int
        Maximum number of mismatches in the barcode.
    
    max_clusters: int
        Maximum number of clusters.
    signal_thresh: float
        Signal threshold.
    info_thresh: float
        Information threshold.
    include_g_u: bool
        Include G-U pairs.
    inlcude_del: bool
        Include deletions.
    min_reads: int
        Minimum number of reads to cluster a reference.
    convergence_cutoff: float
        Convergence cutoff.
    min_iterations: int
        Minimum number of iterations.
    num_runs: int
        Number of runs.
    n_cpus: int
        Number of CPUs.
    
    rnastructure_path: str
        Path to RNAstructure, to predict structure and free energy.
    rnastructure_temperature: bool
        Use samples.csv temperature values for RNAstructure.
    rnastructure_fold_args: str
        Options to pass to RNAstructure fold.
    rnastructure_dms: bool
        Use the DMS signal to make predictions with RNAstructure.
    rnastructure_dms_min_unpaired_value: int
        Minimum unpaired value for using the dms signal as an input for RNAstructure.
    rnastructure_dms_max_paired_value: int
        Maximum paired value for using the dms signal as an input for RNAstructure.
    rnastructure_partition: bool
        Use RNAstructure partition function to predict free energy.
    rnastructure_probability: bool
        Use RNAstructure partition function to predict per-base mutation probability.
    poisson: bool
        Predict Poisson confidence intervals.
    
    verbose: bool
        Verbose output.
    
    """
    def verbose_print(*args):
        print(*args) if verbose else None

    # make output and temp folders
    for folder in ['output', 'temp']:
        os.makedirs(os.path.join(out_dir, folder), exist_ok=True)
        
    # sort fast pairs
    fastq1, fastq2, samples_names = util.util.sort_fastq_pairs(fastq1, fastq2)
    
    # Run DREEM
    verbose_print("""

    ========================================

                RUNNING   DREEM

    ========================================

    """)
    # -----------------------------------------------------------------------------------------------------------------------
    
    ## Demultiplexing (optional)
    # -----------------------------------------------------------------------------------------------------------------------
    if demultiplexing:
        verbose_print('\ndemultiplexing \n------------------')
        
        dreem.demultiplexing.run(out_dir = os.path.join(out_dir, 'output', 'demultiplexing'),
                           fastq1 = fastq1, 
                           fastq2 = fastq2,
                           fasta = fasta,
                           library=library,
                           interleaved=interleaved,
                           barcode_length = barcode_length,
                           barcode_start = barcode_start,
                           max_barcode_mismatches = max_barcode_mismatches)
        verbose_print('demultiplexing done')
    # -----------------------------------------------------------------------------------------------------------------------
    
    ## Alignment: 
    # -----------------------------------------------------------------------------------------------------------------------
    verbose_print('\nalignment \n----------------')
    if demultiplexing:
        fastq1, fastq2, samples_names = [], [], []
        for sample in os.listdir(os.path.join(out_dir, 'output', 'demultiplexing')):
            path = os.path.join(os.path.join(out_dir, 'output', 'demultiplexing'), sample)
            fastq1 = fastq1 + [os.path.join(path, f) for f in os.listdir(path) if f.endswith('_R1.fastq')]
            fastq2 = fastq2 + [os.path.join(path, f) for f in os.listdir(path) if f.endswith('_R2.fastq')]
            samples_names.append(samples_names)

    for idx,(f1, f2, sample) in enumerate(zip(fastq1, fastq2, samples_names)):
        verbose_print('Aligning this fastq pair: ', '\n   ',f1, '\n   ',f2)
        dreem.alignment.run(
                        out_dir=os.path.join(out_dir),#, 'output','alignment'),
                        fasta=fasta,
                        fastq=f1,
                        fastq2=f2,
                        demultiplexed=demultiplexing
                        )
    # -----------------------------------------------------------------------------------------------------------------------
    
    ## Vectoring
    # -----------------------------------------------------------------------------------------------------------------------
    verbose_print('\nvectoring \n------------------')
    for sample in samples_names:
        path_to_bam = os.path.join(out_dir, 'output', 'alignment', sample)
        bam_files = [os.path.join(path_to_bam, f) for f in os.listdir(path_to_bam) if f.endswith('.bam')]
        dreem.vectoring.run(
                       out_dir= os.path.join(out_dir, 'output'), #TODO
                       bam_dirs= bam_files,
                       fasta=fasta,
                       coords=coords,
                       primers=primers,
                       fill=fill,
                       library=library,
                       parallel=parallel
                       )
    # -----------------------------------------------------------------------------------------------------------------------

    ## Clustering (optional)
    # -----------------------------------------------------------------------------------------------------------------------
    if clustering:
        verbose_print('\nclustering \n------------------')
        
        for sample in samples_names:
            dreem.clustering.run(
                           input_dir = os.path.join(out_dir, 'output', 'vectoring'), 
                           out_dir = os.path.join(out_dir, 'output', 'clustering'),
                           max_clusters = max_clusters,
                           min_iter = min_iter,
                           signal_thresh = signal_thresh,
                           info_thresh = info_thresh,
                           include_del = include_del,
                           include_g_u = include_g_u,
                           min_reads = min_reads,
                           convergence_cutoff = convergence_cutoff,
                           num_runs = num_runs,
                           n_cpus = n_cpus
                           )
    # -----------------------------------------------------------------------------------------------------------------------

    ## Aggregate
    # -----------------------------------------------------------------------------------------------------------------------
    verbose_print('\naggregation \n------------------')
    for sample in samples_names:
        clustering_file = os.path.join(out_dir, 'output', 'clustering', sample+'.json') if clustering else None
        vect_out_dir = os.path.join(out_dir, 'output', 'vectoring', sample)
        constructs = [c for c in os.listdir(vect_out_dir) if os.path.isdir(os.path.join(vect_out_dir, c))]
        sections = [[s for s in os.listdir(os.path.join(vect_out_dir, c)) if os.path.isdir(os.path.join(vect_out_dir, c, s))] for c in constructs]
        dreem.aggregation.run(
                        bv_files= [os.path.join(out_dir,'output','vectoring', sample, c, s) for c in constructs for s in sections[constructs.index(c)]],
                        fasta=fasta,
                        library= library,
                        sample = sample, 
                        samples= samples,
                        clustering_file=clustering_file,
                        out_dir= os.path.join(out_dir, 'output', 'aggregation'),
                        rnastructure_path=rnastructure_path,
                        rnastructure_temperature=rnastructure_temperature,
                        rnastructure_fold_args=rnastructure_fold_args,
                        rnastructure_dms=rnastructure_dms,
                        rnastructure_dms_max_paired_value=rnastructure_dms_max_paired_value,
                        rnastructure_dms_min_unpaired_value=rnastructure_dms_min_unpaired_value,
                        rnastructure_partition=rnastructure_partition,
                        rnastructure_probability=rnastructure_probability,
                        poisson=poisson,
                        coords=coords,
                        primers=primers,
                        fill=fill)

    verbose_print('Done!')
    # -----------------------------------------------------------------------------------------------------------------------
